searchData={"items":[{"type":"task","title":"mix skitter.master","doc":"Start a Skitter master node\n\nThis task starts a Skitter master node. It accepts a single option (`--deploy`) and multiple\narguments, which represent worker nodes to connect to. In order to connect to the specified\nworkers, additional arguments need to be passed to the `elixir` or `iex` command used to start\nthe system. More information can be found in the \"Distribution Parameters\" section below.\n\nIt is not recommended to use this task in production. Consider building a release (as described\non the [deployment page](deployment.html#releases)) instead. If mix is used anyway, be sure to\nstart in production mode.","ref":"Mix.Tasks.Skitter.Master.html"},{"type":"task","title":"Options and Arguments - mix skitter.master","doc":"* `--deploy`, `-d`: Deploy the workflow returned by the expression.\n\nBesides the `--deploy` option, any other argument is interpreted as the name of a worker node.\nThe master node will attempt to connect to all the specified nodes when it is starting. If this\nfails, the master exits.","ref":"Mix.Tasks.Skitter.Master.html#module-options-and-arguments"},{"type":"task","title":"Distribution Parameters - mix skitter.master","doc":"In order to connect with other Skitter nodes, the local node needs to be distributed. This task\ndoes not handle distribution, instead, the correct parameters should be passed to `iex` or\n`elixir`. A few examples are provided below:\n\n- `elixir --sname master -S mix skitter.master`\n- `iex --sname master -S mix skitter.master`.\n- `elixir --name master@hostname -S mix skitter.master`\n- `iex --name master@hostname -S mix skitter.master`.","ref":"Mix.Tasks.Skitter.Master.html#module-distribution-parameters"},{"type":"task","title":"mix skitter.new","doc":"Create a Skitter project. It expects the path of the project as an argument.\n\n    $ mix skitter.new PATH [--module MODULE] [--no-fetch-deps] [--fetch-deps] [--develop]\n\nThis task creates a new Skitter project, similar to `mix new` and `mix phx.new`. The project is\ncreated at PATH and is set up to allow the use of Skitter.\n\nSpecifically, the generated project will contain the following files:\n\n* a `mix.exs` file which declares Skitter as a dependency and which is set up to build a\n  Skitter release.\n\n* a `config/config.exs` file which provides some initial configuration for Skitter and logging.\n\n* a `lib/MODULE.ex` file which contain an empty workflow to get started.\n\n* a `.formatter.exs` file which sets up `mix format`\n\n* a `.iex.exs` file which auto-imports Skitter when starting `iex`.\n\n* a `README.md` file which documents the project layout.\n\n* a `.gitignore` file.\n\nAfter setting up the project file, this task will fetch the project dependencies if required.","ref":"Mix.Tasks.Skitter.New.html"},{"type":"task","title":"Flags and arguments - mix skitter.new","doc":"* `--develop`: use the latest development version of Skitter (from github) instead of the\n  current stable version (from hex.pm).\n\n* `--module MODULE`: specify the name of the generated module. If this is not provided, a name\n  is generated based on PATH.\n\n* `--(no-)deps`: By default, the installer asks the user if it should run `mix deps.get` and\n`mix deps.compile` after creating the project. Set this flag to skip this prompt and to\nexplicitly opt in or out of automatically running these commands.","ref":"Mix.Tasks.Skitter.New.html#module-flags-and-arguments"},{"type":"task","title":"mix skitter.worker","doc":"Start a Skitter worker node\n\nThis task starts a Skitter worker node. It accepts a single argument which represents the master\nthe worker will try to connect to and several options.  In order to be able to connect to the\nspecified master (and other workers), additional arguments need to be passed to the `elixir` or\n`iex` command used to start the system. More information can be found in the \"Distribution\nParameters\" section below.\n\nIt is not recommended to use this task in production. Consider using the `skitter.release` task\nto build a release instead. If mix is used anyway, be sure to start in production mode.","ref":"Mix.Tasks.Skitter.Worker.html"},{"type":"task","title":"Flags and Arguments - mix skitter.worker","doc":"* `--no-shutdown-with-master`: By default, a worker node shuts itself down when its connected\nmaster node disconnects. This option can be passed to override this behaviour.\n* `--tag` or `-t`: Specify a `t:Skitter.Nodes.tag/0` for this worker node. This option can be\nused multiple times.\n\nBesides this, the name of a master node can be passed as an argument. The worker will attempt to\nconnect to the specified master. If this is not successful, a warning is logged, however, the\nworker will not shut down.","ref":"Mix.Tasks.Skitter.Worker.html#module-flags-and-arguments"},{"type":"task","title":"Distribution Parameters - mix skitter.worker","doc":"In order to connect with other Skitter nodes, the local node needs to be distributed. This task\ndoes not handle distrbution, instead, the correct parameters should be passed to `iex` or\n`elixir`. A few examples are provided below:\n\n- `elixir --sname worker -S mix skitter.worker`\n- `iex --sname worker -S mix skitter.worker`.\n- `elixir --name worker@hostname -S mix skitter.worker`\n- `iex --name worker@hostname -S mix skitter.worker`.","ref":"Mix.Tasks.Skitter.Worker.html#module-distribution-parameters"},{"type":"module","title":"Skitter","doc":"Skitter framework DSLs and runtime interaction.\n\nThis module serves as an entry point to access both the DSLs implemented by Skitter and to\ninteract with the Skitter runtime system. The former is done by adding `use Skitter` to an\nElixir file or module, while the latter is done by calling the functions defined in this module.","ref":"Skitter.html"},{"type":"module","title":"Domain-specific languages - Skitter","doc":"Adding `use Skitter.DSL` to an Elixir file or module imports the following macros:\n- `Skitter.DSL.Operation.defoperation/3`\n- `Skitter.DSL.Strategy.defstrategy/3`\n- `Skitter.DSL.Workflow.workflow/2`\n\nWhich can be used to define operations, strategies and workflows.","ref":"Skitter.html#module-domain-specific-languages"},{"type":"module","title":"Runtime interaction - Skitter","doc":"The Skitter runtime system is responsible for deploying workflows over a cluster and for\ninteracting with running workflows. The `Skitter.Runtime` module offers an API to interact with\nthe runtime system and the information it tracks. This module offers a shorthand to access the\nmost commonly used functions of this API.","ref":"Skitter.html#module-runtime-interaction"},{"type":"function","title":"Skitter.deploy/1","doc":"Deploy a workflow over the cluster.\n\nStarts a Skitter application (i.e. a `t:Skitter.Workflow.t/0`) by deploying it over the cluster.\nThe workflow is flattened (using `Skitter.Workflow.flatten/1`) before it is deployed. After\ndeployment, this function returns a `t:Skitter.Runtime.ref/0`, which can be used by the\nfunctions in `Skitter.Runtime`, or to stop the workflow.","ref":"Skitter.html#deploy/1"},{"type":"function","title":"Skitter.stop/1","doc":"Stop a deployed workflow with reference `ref`.","ref":"Skitter.html#stop/1"},{"type":"module","title":"Skitter.DSL.Operation","doc":"Callback and Operation definition DSL.\n\nThis module offers macros to define operation modules and callbacks. An operation is defined\nthrough the use of the `defoperation/3` macro. The other macros defined in this module are meant\nto be used inside the body of this macro. We recommend reading the documentation for\n`defoperation/3` first.","ref":"Skitter.DSL.Operation.html"},{"type":"macro","title":"Skitter.DSL.Operation.<~/2","doc":"Updates the current state inside a callback.\n\nThis macro should only be used inside the body of `defcb/2`. It updates the current value of the\noperation state to the provided value.\n\nThis macro can be used  in two ways: it can be used to update the operation state or a field of\nthe operation state. The latter option can only be used if the state of the operation is a\nstruct (i.e. if the intial state has been defined using `state_struct/1`). The former options\nmodifies the operation state as a whole, the second option only modifies the value of the\nprovided field stored in the operation state.","ref":"Skitter.DSL.Operation.html#%3C~/2"},{"type":"macro","title":"Examples - Skitter.DSL.Operation.<~/2","doc":"```\ndefoperation WriteExample do\n  defcb write(), do: state <~ :foo\nend\n```\n    iex> Operation.call(WriteExample, :write, nil, nil, []).state\n    :foo\n\n```\ndefoperation FieldWriteExample do\n  state_struct [:field]\n  defcb write(), do: field <~ :bar\nend\n```\n    iex> Operation.call(FieldWriteExample, :write, %FieldWriteExample{field: :foo}, nil, []).state.field\n    :bar\n\n```\ndefoperation WrongFieldWriteExample do\n  fields [:field]\n  defcb write(), do: doesnotexist <~ :bar\nend\n```\n    iex> Operation.call(WrongFieldWriteExample, :write, %WrongFieldWriteExample{field: :foo}, nil, [])\n    ** (KeyError) key :doesnotexist not found in: %Skitter.DSL.OperationTest.WrongFieldWriteExample{field: :foo}","ref":"Skitter.DSL.Operation.html#%3C~/2-examples"},{"type":"macro","title":"Skitter.DSL.Operation.config/0","doc":"Obtain the operation configuration inside a callback.\n\nThis macro reads the current value of the configuration passed to the operation callback when\nit was called. It should only be used inside the body of `defcb/2`.","ref":"Skitter.DSL.Operation.html#config/0"},{"type":"macro","title":"Examples - Skitter.DSL.Operation.config/0","doc":"```\ndefoperation ConfigExample do\n  defcb read(), do: config()\nend\n```\n\n    iex> Operation.call(ConfigExample, :read, :state, :config, []).result\n    :config","ref":"Skitter.DSL.Operation.html#config/0-examples"},{"type":"macro","title":"Skitter.DSL.Operation.defcb/2","doc":"Define a callback.\n\nThis macro is used to define a callback function. Using this macro, a callback can be defined\nsimilar to a regular procedure. Inside the body of the procedure, `~>/2`, `~>>/2` `<~/2` and\n`sigil_f/2` can be used to access the state and to emit output. `meta_of/1` and `port_of/1` can\nbe used to obtain meta-information about an argument passed to the callback, while\n`inherit_meta/2` and `extend_meta/3` can be used to generate a `Skitter.Token` based to emit\nbased on this meta-information.\n\nInternally, this macro generates a regular elixir function which implements a Skitter callback,\nas defined in `Skitter.Operation`. The macro ensures the generated function returns a\n`t:Skitter.Operation.result/0` with the correct state (as updated by `<~/2`), emit (as updated\nby `~>/2` and `~>>/2`) and result (which contains the value of the last expression in `body`).\nIt also ensures the appropriate meta-information about the callback is added to the operation.\n\nSince `defcb/2` generates a regular Elixir function, pattern matching may still be used in the\nargument list of the callback. Attributes such as `@doc` may also be used as usual.","ref":"Skitter.DSL.Operation.html#defcb/2"},{"type":"macro","title":"Examples - Skitter.DSL.Operation.defcb/2","doc":"```\ndefoperation CbExample do\n  defcb simple(), do: nil\n  defcb arguments(arg1, arg2), do: arg1 + arg2\n  defcb state(), do: counter <~ (~f{counter} + 1)\n  defcb emit_single(), do: ~D[1991-12-08] ~> out_port\n  defcb emit_multi(), do: [~D[1991-12-08], ~D[2021-07-08]] ~>> out_port\nend\n```\n\n    iex> Operation.callbacks(CbExample)\n    MapSet.new([arguments: 2, emit_multi: 0, emit_single: 0, simple: 0, state: 0])\n\n    iex> Operation.callback_info(CbExample, :simple, 0)\n    %Info{read?: false, write?: false, emit?: false}\n\n    iex> Operation.callback_info(CbExample, :arguments, 2)\n    %Info{read?: false, write?: false, emit?: false}\n\n    iex> Operation.callback_info(CbExample, :state, 0)\n    %Info{read?: true, write?: true, emit?: false}\n\n    iex> Operation.callback_info(CbExample, :emit_single, 0)\n    %Info{read?: false, write?: false, emit?: true}\n\n    iex> Operation.callback_info(CbExample, :emit_multi, 0)\n    %Info{read?: false, write?: false, emit?: true}\n\n    iex> Operation.call(CbExample, :simple, %{}, nil, [])\n    %Result{result: nil, emit: [], state: %{}}\n\n    iex> Operation.call(CbExample, :arguments, %{}, nil, [10, 20])\n    %Result{result: 30, emit: [], state: %{}}\n\n    iex> Operation.call(CbExample, :state, %{counter: 10, other: :foo}, nil, [])\n    %Result{result: nil, emit: [], state: %{counter: 11, other: :foo}}\n\n    iex> Operation.call(CbExample, :emit_single, %{}, nil, [])\n    %Result{result: nil, emit: [out_port: [~D[1991-12-08]]], state: %{}}\n\n    iex> Operation.call(CbExample, :emit_multi, %{}, nil, [])\n    %Result{result: nil, emit: [out_port: [~D[1991-12-08], ~D[2021-07-08]]], state: %{}}","ref":"Skitter.DSL.Operation.html#defcb/2-examples"},{"type":"macro","title":"Mutable state and control flow - Skitter.DSL.Operation.defcb/2","doc":"The `<~/2`, `~>/2` and `~>>/2` operators add mutable state to Elixir, which is an immutable\nlanguage. Internally, hidden variables are used to track the current state and values to emit.\nTo make this work, the callback DSL rewrites several control flow structures offered by elixir.\nWhile this approach works, some limitations are present when the `<~/2`, `~>/2` and `~>>/2`\noperators are used inside control flow structures.\n\nThe use of these operators is supported in the following control flow constructs:\n\n* `Kernel.if/2`\n* `Kernel.unless/2`\n* `Kernel.SpecialForms.try/1`\n* `Kernel.SpecialForms.cond/1`\n* `Kernel.SpecialForms.case/2`\n* `Kernel.SpecialForms.receive/1`\n\nWith the exception of `Kernel.SpecialForms.try/1`, all of these control flow constructs behave\nas expected. Said otherwise, state updates performed inside any of these constructs are\nreflected outside of the control flow construct.\n\n### `try`\n\nThe behaviour of `try` blocks is not as straightforward due to implementation limitations. We\ndescribe the different behaviours that may occur below.\n\n```\ndefcb simple(fun) do\n  try do\n    pre <~ :modified\n    res = fun.()\n    post <~ :modified\n    :emit ~> out\n    res\n  rescue\n    RuntimeError ->\n      x <~ :modified\n      :rescue\n  catch\n    _ ->\n      :emit ~> out\n      :catch\n  end\nend\n```\n\nIf no exception is raised or no value is thrown, updates performed inside the `do` block will\nalways be visible:\n\n    iex> Operation.call(TryExample, :simple, %{pre: nil, post: nil}, nil, [fn -> :foo end])\n    %Result{state: %{pre: :modified, post: :modified}, result: :foo, emit: [out: [:emit]]}\n\nHowever, when an exception is raised or a value is thrown, any updates performed inside the `do`\nblock are discarded:\n\n    iex> Operation.call(TryExample, :simple, %{pre: nil, post: nil, x: nil}, nil, [fn -> raise RuntimeError end])\n    %Result{state: %{pre: :nil, post: :nil, x: :modified}, result: :rescue, emit: []}\n\n    iex> Operation.call(TryExample, :simple, %{pre: nil, post: nil}, nil, [fn -> throw :foo end])\n    %Result{state: %{pre: :nil, post: :nil}, result: :catch, emit: [out: [:emit]]}\n\nAs shown above, updates inside the `catch` or `rescue` clauses are reflected in the final\nresult.\n\n#### `else` and `after`\n\n```\ndefcb with_else(fun) do\n  try do\n    pre <~ :modified\n    res = fun.()\n    post <~ :modified\n    :emit ~> out\n    res\n  rescue\n    RuntimeError ->\n      :rescue\n  else\n    res ->\n      x <~ :modified\n      res\n  end\nend\n```\n\nUpdates inside the `do` block _are_ reflected in the `else` block:\n\n    iex> Operation.call(TryExample, :with_else, %{pre: nil, post: nil, x: nil}, nil, [fn -> :foo end])\n    %Result{state: %{pre: :modified, post: :modified, x: :modified}, result: :foo, emit: [out: [:emit]]}\n\nUpdates inside the `after` block are ignored:\n\n```\ndefcb with_after(fun) do\n  try do\n    pre <~ :modified\n    res = fun.()\n    post <~ :modified\n    :emit ~> out\n    res\n  rescue\n    RuntimeError -> :rescue\n  after\n    x <~ :modified\n    :ignored\n  end\nend\n```\n\n    iex> Operation.call(TryExample, :with_after, %{pre: nil, post: nil, x: nil}, nil, [fn -> :foo end])\n    %Result{state: %{pre: :modified, post: :modified, x: nil}, result: :foo, emit: [out: [:emit]]}\n\nIn short, updates are preserved during the normal flow of an operation (i.e. when no values are\nraised or thrown), updates inside `after` are ignored.","ref":"Skitter.DSL.Operation.html#defcb/2-mutable-state-and-control-flow"},{"type":"macro","title":"Skitter.DSL.Operation.defoperation/3","doc":"Define an operation.\n\nThis macro is used to define an operation. Internally, it generates an Elixir module which\nimplements the interface described in `Skitter.Operation`. Inside the body of this macro,\n`initial_state/1` and `state_struct/1` may be used to define the initial state of the operation,\nwhile `defcb/2` may be used to define the various callbacks of the operation.","ref":"Skitter.DSL.Operation.html#defoperation/3"},{"type":"macro","title":"Operation strategy and ports - Skitter.DSL.Operation.defoperation/3","doc":"The operation strategy and its in -and out ports can be defined in the header of the operation\ndeclaration as follows:\n\n    iex> defoperation SignatureExample, in: [a, b, c], out: [y, z], strategy: SomeStrategy do\n    ...> end\n    iex> Operation.strategy(SignatureExample)\n    SomeStrategy\n    iex> Operation.in_ports(SignatureExample)\n    [:a, :b, :c]\n    iex> Operation.out_ports(SignatureExample)\n    [:y, :z]\n\nIf an operation has no `in`, or `out` ports, they can be omitted from the operation's header.\nFurthermore, if the operation only has a single `in` or `out` port, the list notation can be\nomitted:\n\n    iex> defoperation PortExample, in: a do\n    ...> end\n    iex> Operation.in_ports(PortExample)\n    [:a]\n    iex> Operation.out_ports(PortExample)\n    []\n\nThe strategy may be omitted. In this case, a strategy _must_ be provided when the defined\noperation is embedded inside a workflow. If this is not done, an error will be raised when the\nworkflow is deployed.","ref":"Skitter.DSL.Operation.html#defoperation/3-operation-strategy-and-ports"},{"type":"macro","title":"Examples - Skitter.DSL.Operation.defoperation/3","doc":"```\ndefoperation Average, in: value, out: current do\n  state_struct total: 0, count: 0\n\n  defcb react(value) do\n    total <~ ~f{total} + value\n    count <~ ~f{count} + 1\n\n    ~f{total} / ~f{count} ~> current\n  end\nend\n```\n\n    iex> Operation.in_ports(Average)\n    [:value]\n    iex> Operation.out_ports(Average)\n    [:current]\n\n\n    iex> Operation.strategy(Average)\n    nil\n\n    iex> Operation.call(Average, :react, %Average{}, nil, [10])\n    %Result{result: nil, emit: [current: [10.0]], state: %Average{count: 1, total: 10}}\n\n    iex> Operation.call(Average, :react, %Average{count: 1, total: 10}, nil, [10])\n    %Result{result: nil, emit: [current: [10.0]], state: %Average{count: 2, total: 20}}","ref":"Skitter.DSL.Operation.html#defoperation/3-examples"},{"type":"macro","title":"Documentation - Skitter.DSL.Operation.defoperation/3","doc":"When writing documentation for an operation, `@operationdoc` can be used instead of the usual\n`@moduledoc`. When this is done, this macro will automatically add additional information about\nthe operation (such at the ports of the operation) to the generated documentation.","ref":"Skitter.DSL.Operation.html#defoperation/3-documentation"},{"type":"macro","title":"Skitter.DSL.Operation.extend_meta/3","doc":"Create a token with meta-information of an argument and new meta-information inside a callback.\n\nThe provided meta-information overrides any previously existing meta-information.","ref":"Skitter.DSL.Operation.html#extend_meta/3"},{"type":"macro","title":"Examples - Skitter.DSL.Operation.extend_meta/3","doc":"iex> defoperation ExtendMetaExample, in: value do\n    ...>   defcb extend(_value), do: extend_meta(\"hello\", _value, a: 1, b: 2)\n    ...> end\n    iex> Skitter.Operation.call(ExtendMetaExample, :extend, nil, nil, [%Skitter.Token{value: 5, meta: %{foo: :bar}}]).result\n    %Skitter.Token{value: \"hello\", meta: %{a: 1, b: 2, foo: :bar}}\n    iex> Skitter.Operation.call(ExtendMetaExample, :extend, nil, nil, [%Skitter.Token{value: 5, meta: %{a: -1}}]).result\n    %Skitter.Token{value: \"hello\", meta: %{a: 1, b: 2}}\n    iex> Skitter.Operation.call(ExtendMetaExample, :extend, nil, nil, [5]).result\n    %Skitter.Token{value: \"hello\", meta: %{a: 1, b: 2}}","ref":"Skitter.DSL.Operation.html#extend_meta/3-examples"},{"type":"macro","title":"Skitter.DSL.Operation.inherit_meta/2","doc":"Create a new token with the meta-information associated with an argument inside a callback.","ref":"Skitter.DSL.Operation.html#inherit_meta/2"},{"type":"macro","title":"Examples - Skitter.DSL.Operation.inherit_meta/2","doc":"iex> defoperation InheritMetaExample, in: value do\n    ...>   defcb inherit(_value), do: inherit_meta(\"hello\", _value)\n    ...> end\n    iex> Skitter.Operation.call(InheritMetaExample, :inherit, nil, nil, [%Skitter.Token{value: 5, meta: %{foo: :bar}}]).result\n    %Skitter.Token{value: \"hello\", meta: %{foo: :bar}}\n    iex> Skitter.Operation.call(InheritMetaExample, :inherit, nil, nil, [5]).result\n    %Skitter.Token{value: \"hello\", meta: %{}}","ref":"Skitter.DSL.Operation.html#inherit_meta/2-examples"},{"type":"macro","title":"Skitter.DSL.Operation.initial_state/1","doc":"Defines the initial state of an operation.\n\nSkitter operations often deal with state. By convention, the initial state of an operation is\ndefined by a callback named `initial_state`:\n\n```\ndefoperation InitialState do\n  defcb initial_state, do: 0\nend\n```\n\nOften, this callback simply returns a constant value. In this case, this macro can be used as\na shorthand:\n\n```\ndefoperation InitialState do\n  initial_state 0\nend\n```\n\nThe second example generates the code shown in the first example.","ref":"Skitter.DSL.Operation.html#initial_state/1"},{"type":"macro","title":"Examples - Skitter.DSL.Operation.initial_state/1","doc":"iex> Operation.initial_state(InitialState)\n    0","ref":"Skitter.DSL.Operation.html#initial_state/1-examples"},{"type":"macro","title":"Skitter.DSL.Operation.meta_of/1","doc":"Obtain the meta-information associated with an argument inside a callback.","ref":"Skitter.DSL.Operation.html#meta_of/1"},{"type":"macro","title":"Examples - Skitter.DSL.Operation.meta_of/1","doc":"iex> defoperation ReadMetaExample, in: value do\n    ...>   defcb get(_value), do: meta_of(_value)\n    ...> end\n    iex> Skitter.Operation.call(ReadMetaExample, :get, nil, nil, [%Skitter.Token{value: 5, meta: %{foo: :bar}}]).result\n    %{foo: :bar}\n    iex> Skitter.Operation.call(ReadMetaExample, :get, nil, nil, [5]).result\n    %{}","ref":"Skitter.DSL.Operation.html#meta_of/1-examples"},{"type":"macro","title":"Skitter.DSL.Operation.port_of/1","doc":"Obtain the port associated with an argument inside a callback.\n\nIf the argument is not associated with a port, `nil` is returned instead. Usually, the Skitter\nruntime system guarantees the appropriate port is associated with an argument, however, a\nstrategy may prevent port information from being propagated. See `Skitter.Token` for additional\ninformation.","ref":"Skitter.DSL.Operation.html#port_of/1"},{"type":"macro","title":"Examples - Skitter.DSL.Operation.port_of/1","doc":"iex> defoperation PortExample, in: value do\n    ...>   defcb get_port(_value), do: port_of(_value)\n    ...> end\n    iex> Skitter.Operation.call(PortExample, :get_port, nil, nil, [%Skitter.Token{value: 5, port: :value}]).result\n    :value\n    iex> Skitter.Operation.call(PortExample, :get_port, nil, nil, [5]).result\n    nil","ref":"Skitter.DSL.Operation.html#port_of/1-examples"},{"type":"macro","title":"Skitter.DSL.Operation.sigil_f/2","doc":"Read the current value of a field stored in state inside a callback.\n\nThis macro expects that the current operation state is a struct (i.e. it expects an operation\nthat uses `state_struct/1`), and reads the current value of `field` from the struct.\n\nThis macro should only be used inside the body of `defcb/2`.","ref":"Skitter.DSL.Operation.html#sigil_f/2"},{"type":"macro","title":"Examples - Skitter.DSL.Operation.sigil_f/2","doc":"```\ndefoperation FieldReadExample do\n  state_struct field: nil\n  defcb read(), do: ~f{field}\nend\n```\n\n    iex> Operation.call(FieldReadExample, :read, %FieldReadExample{field: 5}, nil, []).result\n    5\n\n    iex> Operation.call(FieldReadExample, :read, %FieldReadExample{field: :foo}, nil, []).result\n    :foo","ref":"Skitter.DSL.Operation.html#sigil_f/2-examples"},{"type":"macro","title":"Skitter.DSL.Operation.state/0","doc":"Obtain the current state inside a callback.\n\nThis macro reads the current value of the state passed to the operation callback when it was\ncalled. It should only be used inside the body of `defcb/2`.","ref":"Skitter.DSL.Operation.html#state/0"},{"type":"macro","title":"Examples - Skitter.DSL.Operation.state/0","doc":"```\ndefoperation ReadExample do\n  initial_state 0\n  defcb read(), do: state()\nend\n```\n\n    iex> Operation.call(ReadExample, :read, :state, nil, []).result\n    :state","ref":"Skitter.DSL.Operation.html#state/0-examples"},{"type":"macro","title":"Skitter.DSL.Operation.state_struct/1","doc":"Creates an initial struct-based state for an operation.\n\nIn Elixir, it is common to use a struct to store structured information. Therefore, when an\noperation manages a complex state, it often defines a struct and uses this struct as the initial\nstate of the operation. Afterwards, the state of the operation is updated when it reacts to\nincoming data:\n\n```\ndefoperation Average, in: value, out: current do\n  defstruct [total: 0, count: 0]\n  initial_state %__MODULE__{}\n\n  defcb react(val) do\n    state <~ %{state() | count: state().count + 1}\n    state <~ %{state() | total: state().total + val}\n    state().total / state().count ~> current\n  end\nend\n```\n\nIn order to streamline the use of this pattern, this macro defines a struct and uses this struct\nas the initial state of the operation. Moreover, the `sigil_f/2` and `<~/2` macros are designed\nto be used with structs, enabling them to read the state and update it:\n\n```\ndefoperation Average, in: value, out: current do\n  state_struct total: 0, count: 0\n\n  defcb react(val) do\n    count <~ ~f{count} + 1\n    total <~ ~f{total} + val\n    ~f{total} / ~f{count} ~> current\n  end\nend\n```\n\nThe second example generates the code shown in the first example.","ref":"Skitter.DSL.Operation.html#state_struct/1"},{"type":"macro","title":"Examples - Skitter.DSL.Operation.state_struct/1","doc":"iex> Operation.initial_state(Average)\n    %Average{total: 0, count: 0}","ref":"Skitter.DSL.Operation.html#state_struct/1-examples"},{"type":"macro","title":"Skitter.DSL.Operation.~>/2","doc":"Emit `value` to `port` inside a callback.\n\nThis macro is used to specify `value` should be emitted on `port`. This means that `value`\nwill be sent to any operations downstream of the current operation. This macro should only be\nused inside the body of `defcb/2`. If a previous value was specified for `port`, it is\noverridden.","ref":"Skitter.DSL.Operation.html#~%3E/2"},{"type":"macro","title":"Examples - Skitter.DSL.Operation.~>/2","doc":"```\ndefoperation SingleEmitExample do\n  defcb emit(value) do\n    value ~> some_port\n    :foo ~> some_other_port\n  end\nend\n```\n\n    iex> Operation.call(SingleEmitExample, :emit, nil, nil, [:bar]).emit\n    [some_other_port: [:foo], some_port: [:bar]]","ref":"Skitter.DSL.Operation.html#~%3E/2-examples"},{"type":"macro","title":"Skitter.DSL.Operation.~>>/2","doc":"Emit several values to `port` inside a callback.\n\nThis macro works like `~>/2`, but emits several output values to the port instead of a single\nvalue. Each value in the provided `t:Enumerable.t/0` will be sent to downstream operations\nindividually.","ref":"Skitter.DSL.Operation.html#~%3E%3E/2"},{"type":"macro","title":"Examples - Skitter.DSL.Operation.~>>/2","doc":"```\ndefoperation MultiEmitExample do\n  defcb emit(value) do\n    value ~> some_port\n    [:foo, :bar] ~>> some_other_port\n  end\nend\n```\n\n    iex> Operation.call(MultiEmitExample, :emit, nil, nil, [:bar]).emit\n    [some_other_port: [:foo, :bar], some_port: [:bar]]","ref":"Skitter.DSL.Operation.html#~%3E%3E/2-examples"},{"type":"module","title":"Skitter.DSL.Strategy","doc":"Strategy and Hook definition DSL.\n\nThis module offers macros to define a strategy and its hooks. A strategy is defined through the\nuse of the `defstrategy/3` macro. Inside the body of `defstrategy/3`, `defhook/2` can be used\nto define a hook. The other macros in this module can be used inside the body of `defhook/2` to\nobtain information from the current `t:Skitter.Strategy.context/0`. We recommend reading the\ndocumentation of `defstrategy/3` to get started.","ref":"Skitter.DSL.Strategy.html"},{"type":"macro","title":"Skitter.DSL.Strategy.context/0","doc":"Obtain the context struct.\n\nA strategy hook is called with a `t:Skitter.Strategy.context/0` as its first argument. This\nmacro is used to obtain this struct.","ref":"Skitter.DSL.Strategy.html#context/0"},{"type":"macro","title":"Examples - Skitter.DSL.Strategy.context/0","doc":"iex> defstrategy FullContext do\n    ...>   defhook read, do: context()\n    ...> end\n    iex> FullContext.read(%Context{operation: SomeOperation})\n    %Context{operation: SomeOperation}","ref":"Skitter.DSL.Strategy.html#context/0-examples"},{"type":"macro","title":"Skitter.DSL.Strategy.defhook/2","doc":"Define a hook.\n\nThis macro defines a strategy hook. Hooks are functions called by the Skitter runtime system in\nresponse to predefined events (described in `Skitter.Strategy.Operation`), or by other hooks.\nInternally, hooks are plain Elixir functions which accept a `t:Skitter.Strategy.context/0` as\ntheir first argument. This macro generates a plain Elixir function which accepts such a context.\n\nSaid otherwise, the following two definitions are equivalent:\n\n```\ndef my_hook(context, arg1, arg2), do: ...\n```\n\n```\ndefhook my_hook(arg1, arg2), do: ...\n```\n\nUsing the `defhook/2` macro, however, ensures hooks can be inherited (as described in\n`defstrategy/3`), and offers access to the macros defined in `Skitter.DSL.Strategy.Helpers`,\nwhich provide the building blocks required to build a strategy. The `context/0`, `operation/0`,\n`strategy/0` and `deployment/0` macros can be used to obtain the information stored in the\ncontext the hook was called with.","ref":"Skitter.DSL.Strategy.html#defhook/2"},{"type":"macro","title":"Calling hooks - Skitter.DSL.Strategy.defhook/2","doc":"Since hooks are plain Elixir functions, they may be called like any other function. However, a\n`t:Skitter.Strategy.context/0` must be provided:\n\n    iex> defstrategy Strategy do\n    ...>   defhook hook, do: \"hello\"\n    ...> end\n    iex> Strategy.hook(%Context{})\n    \"hello\"\n\nWhen a hook calls another hook (defined in another strategy or in the same strategy), it _must_\nuse the `context/0` macro to pass the current context.\n\n    iex> defstrategy S1 do\n    ...>   defhook example, do: \"world!\"\n    ...> end\n    iex> defstrategy S2 do\n    ...>   defhook example, do: \"Hello, \" <> S1.example(context())\n    ...> end\n    iex> S2.example(%Context{})\n    \"Hello, world!\"\n\n    iex> defstrategy Local do\n    ...>   defhook left, do: \"Hello, \"\n    ...>   defhook right, do: \"world!\"\n    ...>   defhook example, do: left(context()) <> right(context())\n    ...> end\n    iex> Local.example(%Context{})\n    \"Hello, world!\"\n\nSince the context contains the current strategy, it is possible to dynamically call the current\nstrategy. However, the syntax for this is rather unwieldy:\n\n    iex> defstrategy AbstractAnimal do\n    ...>   defhook example, do: \"Animal says: \" <> strategy().say(context())\n    ...> end\n    iex> defstrategy Dog, extends: AbstractAnimal do\n    ...>   defhook say, do: \"woof!\"\n    ...> end\n    iex> Dog.example(%Context{strategy: Dog})\n    \"Animal says: woof!\"\n    iex> defstrategy Cat, extends: AbstractAnimal do\n    ...>   defhook say, do: \"meow!\"\n    ...> end\n    iex> Cat.example(%Context{strategy: Cat})\n    \"Animal says: meow!\"\n\nNote that we have to explicitly pass the strategy as a part of the context in the examples\nabove. In a real application, the Skitter Runtime calls hooks, ensuring the appropriate context\nis passed.","ref":"Skitter.DSL.Strategy.html#defhook/2-calling-hooks"},{"type":"macro","title":"Skitter.DSL.Strategy.defstrategy/3","doc":"Define a strategy.\n\nThis macro is used to define a strategy. A `Skitter.Strategy` is defined as a regular Elixir\nmodule which defines several _hooks_. As such, any code that is valid inside an Elixir module\n(such as function definitions or module attributes) is valid inside `defstrategy/3`.\nAdditionally, the `defhook/2` macro may be used to define a Skitter _hook_. Hooks are special\nElixir functions which accept a `t:Skitter.Strategy.context/0` argument.","ref":"Skitter.DSL.Strategy.html#defstrategy/3"},{"type":"macro","title":"Extending Strategies - Skitter.DSL.Strategy.defstrategy/3","doc":"Strategies can be created based on existing strategies. This is done by _extending_ some\nstrategy. When a strategy extends another strategy, it will inherit all the hooks defined by the\nstrategy it extends:\n\n    iex> defstrategy Parent do\n    ...>   defhook example, do: :example_hook\n    ...> end\n    iex> defstrategy Child, extends: Parent do\n    ...> end\n    iex> Child.example(%Context{})\n    :example_hook\n\nInherited hooks can be overridden:\n\n    iex> defstrategy Parent do\n    ...>   defhook example, do: :parent\n    ...> end\n    iex> defstrategy Child, extends: Parent do\n    ...>   defhook example, do: :child\n    ...> end\n    iex> Child.example(%Context{})\n    :child\n\nFinally, a strategy can extend multiple parent strategies. When this is done, the hooks of\nearlier parent strategies take precedence over later hooks:\n\n    iex> defstrategy Parent1 do\n    ...>   defhook example, do: :parent1\n    ...> end\n    iex> defstrategy Parent2 do\n    ...>   defhook example, do: :parent2\n    ...>   defhook another, do: :parent2\n    ...> end\n    iex> defstrategy Child, extends: [Parent1, Parent2] do\n    ...> end\n    iex> Child.example(%Context{})\n    :parent1\n    iex> Child.another(%Context{})\n    :parent2","ref":"Skitter.DSL.Strategy.html#defstrategy/3-extending-strategies"},{"type":"macro","title":"Skitter.DSL.Strategy.deployment/0","doc":"Obtain the context's deployment.","ref":"Skitter.DSL.Strategy.html#deployment/0"},{"type":"macro","title":"Examples - Skitter.DSL.Strategy.deployment/0","doc":"iex> defstrategy ReadDeployment do\n    ...>   defhook read, do: deployment()\n    ...> end\n    iex> ReadDeployment.read(%Context{deployment: :some_deployment_data})\n    :some_deployment_data","ref":"Skitter.DSL.Strategy.html#deployment/0-examples"},{"type":"macro","title":"Skitter.DSL.Strategy.operation/0","doc":"Obtain the context's operation.","ref":"Skitter.DSL.Strategy.html#operation/0"},{"type":"macro","title":"Examples - Skitter.DSL.Strategy.operation/0","doc":"iex> defstrategy ReadOperation do\n    ...>   defhook read, do: operation()\n    ...> end\n    iex> ReadOperation.read(%Context{operation: SomeOperation})\n    SomeOperation","ref":"Skitter.DSL.Strategy.html#operation/0-examples"},{"type":"macro","title":"Skitter.DSL.Strategy.strategy/0","doc":"Obtain the context's strategy.","ref":"Skitter.DSL.Strategy.html#strategy/0"},{"type":"macro","title":"Examples - Skitter.DSL.Strategy.strategy/0","doc":"iex> defstrategy ReadStrategy do\n    ...>   defhook read, do: strategy()\n    ...> end\n    iex> ReadStrategy.read(%Context{strategy: ReadStrategy})\n    Skitter.DSL.StrategyTest.ReadStrategy","ref":"Skitter.DSL.Strategy.html#strategy/0-examples"},{"type":"module","title":"Skitter.DSL.Strategy.Helpers","doc":"Macros to be used in strategy hooks.\n\nThis module defines various macro \"primitives\" which can be used to define strategies. These\nmacros are only available inside `Skitter.DSL.Strategy.defhook/2`, where they are automatically\nimported.\n\nInternally, these macros call functions defined in various other modules. The information stored\nin the `t:Skitter.Strategy.context/0` passed to the hook is used to pass the appropriate\ninformation to these functions.","ref":"Skitter.DSL.Strategy.Helpers.html"},{"type":"macro","title":"Skitter.DSL.Strategy.Helpers.call/2","doc":"Call `callback` of the current operation.\n\nUses `Skitter.Operation.call/5`. The state, configuration and arguments can be passed through\nopts. If they are not provided, `:args` defaults to the empty list while `:state` and  `:config`\ndefault to `nil`.","ref":"Skitter.DSL.Strategy.Helpers.html#call/2"},{"type":"macro","title":"Examples - Skitter.DSL.Strategy.Helpers.call/2","doc":"iex> defoperation ExampleOperation do\n    ...>   defcb example(arg), do: {arg, state(), config()}\n    ...> end\n    iex> defstrategy ExampleStrategy do\n    ...>   defhook empty_example(), do: call(:example).result\n    ...>   defhook arg_example(), do: call(:example, args: [:foo]).result\n    ...>   defhook state_example(), do: call(:example, args: [nil], state: :foo).result\n    ...>   defhook config_example(), do: call(:example, args: [nil], config: :foo).result\n    ...>   defhook all_example(), do: call(:example, args: [:foo], state: :foo, config: :foo).result\n    ...> end\n    iex> ExampleStrategy.arg_example(%Context{operation: ExampleOperation})\n    {:foo, nil, nil}\n    iex> ExampleStrategy.state_example(%Context{operation: ExampleOperation})\n    {nil, :foo, nil}\n    iex> ExampleStrategy.config_example(%Context{operation: ExampleOperation})\n    {nil, nil, :foo}\n    iex> ExampleStrategy.all_example(%Context{operation: ExampleOperation})\n    {:foo, :foo, :foo}\n    iex> ExampleStrategy.empty_example(%Context{operation: ExampleOperation})\n    ** (UndefinedFunctionError) function Skitter.DSL.Strategy.HelpersTest.ExampleOperation.example/2 is undefined or private","ref":"Skitter.DSL.Strategy.Helpers.html#call/2-examples"},{"type":"macro","title":"Skitter.DSL.Strategy.Helpers.call_if_exists/2","doc":"Call `callback` of the current operation if it exists.\n\nUses `Skitter.Operation.call_if_exists/5`. The state, configuration and arguments can be passed\nthrough opts (as in `call/3`). If they are not provided, `:args` defaults to the empty list\nwhile `:state` and `:config` default to `nil`.","ref":"Skitter.DSL.Strategy.Helpers.html#call_if_exists/2"},{"type":"macro","title":"Examples - Skitter.DSL.Strategy.Helpers.call_if_exists/2","doc":"iex> defoperation ExampleOperation, out: port do\n    ...>   defcb exists do\n    ...>     state <~ :exists\n    ...>     :exists ~> port\n    ...>     :exists\n    ...>   end\n    ...> end\n    iex> defstrategy ExampleStrategy do\n    ...>   defhook exists_example(), do: call_if_exists(:exists)\n    ...>   defhook does_not_exist_example(), do: call_if_exists(:does_not_exist)\n    ...> end\n    iex> ExampleStrategy.exists_example(%Context{operation: ExampleOperation})\n    %Result{result: :exists, state: :exists, emit: [port: [:exists]]}\n    iex> ExampleStrategy.does_not_exist_example(%Context{operation: ExampleOperation})\n    %Result{result: nil, state: nil, emit: []}","ref":"Skitter.DSL.Strategy.Helpers.html#call_if_exists/2-examples"},{"type":"macro","title":"Skitter.DSL.Strategy.Helpers.emit/1","doc":"Emit data for the current context.\n\nUses `Skitter.Strategy.Operation.emit/2`","ref":"Skitter.DSL.Strategy.Helpers.html#emit/1"},{"type":"macro","title":"Skitter.DSL.Strategy.Helpers.error/1","doc":"Raise a `Skitter.StrategyError`\n\nThe error is automatically annotated with the current context, which is used to retrieve the\ncurrent operation and strategy.","ref":"Skitter.DSL.Strategy.Helpers.html#error/1"},{"type":"macro","title":"Examples - Skitter.DSL.Strategy.Helpers.error/1","doc":"iex> defstrategy Example do\n    ...>   defhook example, do: error(\"An error message\")\n    ...> end\n    iex> Example.example(%Context{strategy: Example, operation: Foo})\n    ** (Skitter.StrategyError) Raised by Skitter.DSL.Strategy.HelpersTest.Example handling Foo:\n       An error message","ref":"Skitter.DSL.Strategy.Helpers.html#error/1-examples"},{"type":"macro","title":"Skitter.DSL.Strategy.Helpers.in_port_to_index/1","doc":"Get the index of the given in `port`.\n\nCalls `Skitter.Operation.in_port_to_index/2`.","ref":"Skitter.DSL.Strategy.Helpers.html#in_port_to_index/1"},{"type":"macro","title":"Examples - Skitter.DSL.Strategy.Helpers.in_port_to_index/1","doc":"iex> defoperation ExampleOperation, in: [foo, bar] do\n    ...> end\n    iex> defstrategy ExampleStrategy do\n    ...>   defhook example(), do: in_port_to_index(:bar)\n    ...> end\n    iex> ExampleStrategy.example(%Context{operation: ExampleOperation})\n    1","ref":"Skitter.DSL.Strategy.Helpers.html#in_port_to_index/1-examples"},{"type":"macro","title":"Skitter.DSL.Strategy.Helpers.index_to_in_port/1","doc":"Get the name of the in port with the given `index`.\n\nCalls `Skitter.Operation.index_to_in_port/2`.","ref":"Skitter.DSL.Strategy.Helpers.html#index_to_in_port/1"},{"type":"macro","title":"Examples - Skitter.DSL.Strategy.Helpers.index_to_in_port/1","doc":"iex> defoperation ExampleOperation, in: [foo, bar] do\n    ...> end\n    iex> defstrategy ExampleStrategy do\n    ...>   defhook example(), do: index_to_in_port(1)\n    ...> end\n    iex> ExampleStrategy.example(%Context{operation: ExampleOperation})\n    :bar","ref":"Skitter.DSL.Strategy.Helpers.html#index_to_in_port/1-examples"},{"type":"macro","title":"Skitter.DSL.Strategy.Helpers.index_to_out_port/1","doc":"Get the name of the out port with the given `index`.\n\nCalls `Skitter.Operation.index_to_out_port/2`.","ref":"Skitter.DSL.Strategy.Helpers.html#index_to_out_port/1"},{"type":"macro","title":"Examples - Skitter.DSL.Strategy.Helpers.index_to_out_port/1","doc":"iex> defoperation ExampleOperation, out: [foo, bar] do\n    ...> end\n    iex> defstrategy ExampleStrategy do\n    ...>   defhook example(), do: index_to_out_port(0)\n    ...> end\n    iex> ExampleStrategy.example(%Context{operation: ExampleOperation})\n    :foo","ref":"Skitter.DSL.Strategy.Helpers.html#index_to_out_port/1-examples"},{"type":"macro","title":"Skitter.DSL.Strategy.Helpers.initial_state/1","doc":"Create an empty state for the operation.\n\nThis macro creates an initial state for an operation, by using\n`Skitter.Operation.initial_state/2`","ref":"Skitter.DSL.Strategy.Helpers.html#initial_state/1"},{"type":"macro","title":"Examples - Skitter.DSL.Strategy.Helpers.initial_state/1","doc":"iex> defoperation ExampleOperation do\n    ...>   initial_state :initial_state\n    ...> end\n    iex> defstrategy ExampleStrategy do\n    ...>   defhook example(), do: initial_state()\n    ...> end\n    iex> ExampleStrategy.example(%Context{operation: ExampleOperation})\n    :initial_state","ref":"Skitter.DSL.Strategy.Helpers.html#initial_state/1-examples"},{"type":"macro","title":"Skitter.DSL.Strategy.Helpers.local_worker/2","doc":"Create a worker using `Skitter.Worker.create_local/3`.\n\nThis macro creates a local worker, automatically passing the current context.","ref":"Skitter.DSL.Strategy.Helpers.html#local_worker/2"},{"type":"macro","title":"Skitter.DSL.Strategy.Helpers.out_port_to_index/1","doc":"Get the index of the given out `port`.\n\nCalls `Skitter.Operation.out_port_to_index/2`.","ref":"Skitter.DSL.Strategy.Helpers.html#out_port_to_index/1"},{"type":"macro","title":"Examples - Skitter.DSL.Strategy.Helpers.out_port_to_index/1","doc":"iex> defoperation ExampleOperation, out: [foo, bar] do\n    ...> end\n    iex> defstrategy ExampleStrategy do\n    ...>   defhook example(), do: out_port_to_index(:foo)\n    ...> end\n    iex> ExampleStrategy.example(%Context{operation: ExampleOperation})\n    0","ref":"Skitter.DSL.Strategy.Helpers.html#out_port_to_index/1-examples"},{"type":"macro","title":"Skitter.DSL.Strategy.Helpers.plain_send/2","doc":"Send a message to a worker using Elixir's `Kernel.send/2`.\n\n`send/2` and `send/3` defined in this module call `Skitter.Worker.send/2` which will send a\nmessage to a worker, eventually causing its `c:Skitter.Strategy.Operation.process/4` hook to be\ncalled.\n\nIn contrast, this function uses the built-in `Kernel.send/2` of Elixir, which sends a message to\na pid. This is useful when you need to use `Kernel.SpecialForms.receive/1` inside a hook.","ref":"Skitter.DSL.Strategy.Helpers.html#plain_send/2"},{"type":"macro","title":"Skitter.DSL.Strategy.Helpers.remote_worker/3","doc":"Create a worker using `Skitter.Worker.create_remote/4`.\n\nThis macro creates a remote worker, automatically passing the current context.","ref":"Skitter.DSL.Strategy.Helpers.html#remote_worker/3"},{"type":"macro","title":"Skitter.DSL.Strategy.Helpers.send/2","doc":"Send a message to a worker with `Skitter.Worker.send/2`","ref":"Skitter.DSL.Strategy.Helpers.html#send/2"},{"type":"macro","title":"Skitter.DSL.Strategy.Helpers.stop_worker/0","doc":"Stop the current worker using `Skitter.Worker.stop/1`","ref":"Skitter.DSL.Strategy.Helpers.html#stop_worker/0"},{"type":"macro","title":"Skitter.DSL.Strategy.Helpers.stop_worker/1","doc":"Stop the given worker using `Skitter.Worker.stop/1`","ref":"Skitter.DSL.Strategy.Helpers.html#stop_worker/1"},{"type":"macro","title":"Skitter.DSL.Strategy.Helpers.to_all_ports/1","doc":"Programmatically create output for all out ports.\n\nThe data must be wrapped in a list.","ref":"Skitter.DSL.Strategy.Helpers.html#to_all_ports/1"},{"type":"macro","title":"Examples - Skitter.DSL.Strategy.Helpers.to_all_ports/1","doc":"iex> defoperation ExampleOperation, out: [foo, bar] do\n    ...> end\n    iex> defstrategy ExampleStrategy do\n    ...>   defhook example(), do: to_all_ports([1, 2, 3])\n    ...> end\n    iex> ExampleStrategy.example(%Context{operation: ExampleOperation})\n    [foo: [1, 2, 3], bar: [1, 2, 3]]","ref":"Skitter.DSL.Strategy.Helpers.html#to_all_ports/1-examples"},{"type":"macro","title":"Skitter.DSL.Strategy.Helpers.to_port/2","doc":"Programmatically create output for the out port with `index`.\n\nThe data must be wrapped in a list.","ref":"Skitter.DSL.Strategy.Helpers.html#to_port/2"},{"type":"macro","title":"Examples - Skitter.DSL.Strategy.Helpers.to_port/2","doc":"iex> defoperation ExampleOperation, out: [foo, bar] do\n    ...> end\n    iex> defstrategy ExampleStrategy do\n    ...>   defhook example(), do: to_port(0, [1, 2, 3])\n    ...> end\n    iex> ExampleStrategy.example(%Context{operation: ExampleOperation})\n    [foo: [1, 2, 3]]","ref":"Skitter.DSL.Strategy.Helpers.html#to_port/2-examples"},{"type":"module","title":"Skitter.DSL.Workflow","doc":"Workflow definition DSL.\n\nThis module offers macros for the definition of workflows. Workflow are defined through the use\nof `workflow/2`. The other macros defined in this module are meant to be used inside the body\nof `workflow/2`. We recommend reading the documentation of `workflow/2` first.","ref":"Skitter.DSL.Workflow.html"},{"type":"macro","title":"Skitter.DSL.Workflow.node/2","doc":"Generate a single workflow node.\n\nThis macro generates a single node of a workflow. It can only be used inside `workflow/2`. It\naccepts a `t:Skitter.Operation.t/0` or a workflow `t:Skitter.Workflow.t/0` and a list of\noptional options. The provided operation or workflow will be wrapped inside a\n`t:Skitter.Workflow.operation_node/0` or `t:Skitter.Workflow.workflow_node/0`. No links will be\nadded to the generated node.\n\nThree options can be passed when creating a node: `as:`, `args:` and `with:`:\n\n- `as:` defines the name of the node inside the workflow. It can be used to refer to the\nnode when creating links with `~>/2`. If no name is specified, the node macro will generate a\nunique name.\n- `args:` defines the arguments to pass to the node. Note that this is only relevant for\noperation nodes. Arguments passed to workflow nodes are ignored. If no arguments are provided,\nthe arguments of the node defaults to `nil`.\n- `with:` defines the strategy to pass to the node. Note that this is only relevant for\noperation nodes. When a strategy is provided here, it will override the one defined by the\noperation. If no strategy is provided, the strategy specified by the operation will be used. If\nno strategy is specified by the operation, an error will be raised.","ref":"Skitter.DSL.Workflow.html#node/2"},{"type":"macro","title":"Examples - Skitter.DSL.Workflow.node/2","doc":"iex> inner = workflow do\n    ...>   node Example\n    ...>   node Example, as: example_1\n    ...>   node Example, args: :args\n    ...>   node Example, as: example_2, args: :args, with: SomeStrategy\n    ...> end\n    %Skitter.Workflow{\n      in: [],\n      out: [],\n      nodes: %{\n        \"skitter/dsl/workflow_test/example#1\": %Skitter.Workflow.Node.Operation{\n          operation: Example, args: nil, strategy: DefaultStrategy, links: []\n        },\n        example_1:  %Skitter.Workflow.Node.Operation{\n          operation: Example, args: nil, strategy: DefaultStrategy, links: []\n        },\n        \"skitter/dsl/workflow_test/example#2\": %Skitter.Workflow.Node.Operation{\n          operation: Example, args: :args, strategy: DefaultStrategy, links: []\n        },\n        example_2:  %Skitter.Workflow.Node.Operation{\n          operation: Example, args: :args, strategy: SomeStrategy, links: []\n        },\n      }\n    }\n    iex> workflow do\n    ...>   node inner\n    ...>   node inner, as: nested_1\n    ...>   node inner, args: :will_be_ignored, as: nested_2\n    ...> end\n    %Skitter.Workflow{\n      in: [],\n      out: [],\n      nodes: %{\n        \"#nested#1\": %Skitter.Workflow.Node.Workflow{workflow: inner, links: []},\n        nested_1: %Skitter.Workflow.Node.Workflow{workflow: inner, links: []},\n        nested_2: %Skitter.Workflow.Node.Workflow{workflow: inner, links: []}\n      }\n    }\n\n    iex> workflow do\n    ...>   node Join\n    ...> end\n    ** (Skitter.DefinitionError) Operation Elixir.Skitter.DSL.WorkflowTest.Join does not define a strategy and no strategy was specified by the workflow\n\n    iex> workflow do\n    ...>   node Join, with: SomeStrategy\n    ...> end\n    %Skitter.Workflow{\n      in: [],\n      out: [],\n      nodes: %{\n        \"skitter/dsl/workflow_test/join#1\": %Skitter.Workflow.Node.Operation{\n          operation: Join, args: nil, strategy: SomeStrategy, links: []\n        }\n      }\n    }","ref":"Skitter.DSL.Workflow.html#node/2-examples"},{"type":"macro","title":"Skitter.DSL.Workflow.workflow/2","doc":"Define a workflow.\n\nThis macro generates a `t:Skitter.Workflow.t/0`. Inside the body of this macro, `node/2` and\n`~>/2` can be used to define nodes and links between nodes, respectively. The generated workflow\nis verified after its definition through the use of `Skitter.Workflow.verify/1`.\n\nInternally, this macro generates the data structure defined in `t:Skitter.Workflow.t/0`.","ref":"Skitter.DSL.Workflow.html#workflow/2"},{"type":"macro","title":"Workflow ports - Skitter.DSL.Workflow.workflow/2","doc":"The ports of a workflow can be defined in the header of the workflow macro as follows:\n\n    iex> wf = workflow in: [a], out: [x] do\n    ...> end\n    iex> wf.in\n    [a: []]\n    iex> wf.out\n    [:x]\n\nIf a workflow has no `in`, or `out` ports, they can be omitted from the workflow header.\nFurthermore, if the workflow only has a single `in` or `out` port, the list notation can be\nomitted:\n\n    iex> wf = workflow in: a do\n    ...> end\n    iex> wf.in\n    [a: []]\n    iex> wf.out\n    []","ref":"Skitter.DSL.Workflow.html#workflow/2-workflow-ports"},{"type":"macro","title":"Nodes, links and syntactic sugar - Skitter.DSL.Workflow.workflow/2","doc":"Inside the body of a workflow, the `node/2` and `~>/2` macros are used to define nodes and to\nlink them to one another:\n\n    iex> wf = workflow do\n    ...>   node Example, as: node1\n    ...>   node Example, as: node2\n    ...>\n    ...>   node1.out_port ~> node2.in_port\n    ...> end\n    iex> wf.nodes[:node1].operation\n    Example\n    iex> wf.nodes[:node1].links\n    [out_port: [node2: :in_port]]\n\n![](assets/docs/dsl_workflow_basic.png)\n\nTo link nodes to the in or out ports of a workflow, the port name should be used:\n\n    iex> wf = workflow in: foo, out: bar do\n    ...>   node Example, as: node\n    ...>\n    ...>   foo ~> node.in_port\n    ...>   node.out_port ~> bar\n    ...> end\n    iex> wf.nodes[:node].links\n    [out_port: [:bar]]\n    iex> wf.in\n    [foo: [node: :in_port]]\n\n![](assets/docs/dsl_workflow_ports.png)\n\nPreviously defined workflows may be used inside a workflow definition:\n\n    iex> inner = workflow in: foo, out: bar do\n    ...>   node Example, as: node\n    ...>\n    ...>   foo ~> node.in_port\n    ...>   node.out_port ~> bar\n    ...> end\n    iex> outer = workflow do\n    ...>   node inner, as: inner_left\n    ...>   node inner, as: inner_right\n    ...>\n    ...>   inner_left.bar ~> inner_right.foo\n    ...> end\n    iex> outer.nodes[:inner_left].workflow == inner\n    true\n    iex> outer.nodes[:inner_left].links\n    [bar: [inner_right: :foo]]\n\n![](assets/docs/dsl_workflow_nested.png)\n\nInstead of specifying the complete source name (e.g. `node.in_port`), the following syntactic\nsugar can be used when creating a node:\n\n    iex> wf = workflow in: foo do\n    ...>   foo ~> node(Example, as: node)\n    ...> end\n    iex> wf.in\n    [foo: [node: :in_port]]\n\n![](assets/docs/dsl_workflow_sugar_1.png)\n\n    iex> wf = workflow out: bar do\n    ...>   node(Example, as: node) ~> bar\n    ...> end\n    iex> wf.nodes[:node].links\n    [out_port: [:bar]]\n\n![](assets/docs/dsl_workflow_sugar_2.png)\n\nThese uses of `~>/2` can be chained:\n\n    iex> wf = workflow in: foo, out: bar do\n    ...>   foo\n    ...>   ~> node(Example, as: node1)\n    ...>   ~> node(Example, as: node2)\n    ...>   ~> bar\n    ...> end\n    iex> wf.in\n    [foo: [node1: :in_port]]\n    iex> wf.nodes[:node1].links\n    [out_port: [node2: :in_port]]\n    iex> wf.nodes[:node2].links\n    [out_port: [:bar]]\n\n![](assets/docs/dsl_workflow_sugar_3.png)\n\nIt is not needed to explicitly specify a name for a node if you do not need to refer to the\nnode. You should not rely on the format of the generated names in this case:\n\n    iex> wf = workflow in: foo, out: bar do\n    ...>   foo\n    ...>   ~> node(Example)\n    ...>   ~> node(Example)\n    ...>   ~> bar\n    ...> end\n    iex> wf.in\n    [foo: [\"skitter/dsl/workflow_test/example#1\": :in_port]]\n    iex> wf.nodes[:\"skitter/dsl/workflow_test/example#1\"].links\n    [out_port: [\"skitter/dsl/workflow_test/example#2\": :in_port]]\n    iex> wf.nodes[:\"skitter/dsl/workflow_test/example#2\"].links\n    [out_port: [:bar]]\n\n![](assets/docs/dsl_workflow_gen_name.png)\n\nSkitter defines several commonly-used operations such as `Skitter.BIO.Map` or\n`Skitter.BIO.Filter`. These operations are all prefixed with `Skitter.BIO`.\n\nThe `Skitter.BIO` module defines several _operators_ which provide nicer syntax for using these\nbuilt-in operations inside a workflow. These operators are automatically imported by\n`workflow/2`.  Thus, the following two workflow definitions are equivalent:\n\n    iex> map_fun = fn i -> i * 2 end\n    iex> without_sugar = workflow do\n    ...>   node(Skitter.BIO.StreamSource, args: [1, 2, 3])\n    ...>   ~> node(Skitter.BIO.Map, args: map_fun)\n    ...>   ~> node(Skitter.BIO.Print)\n    ...> end\n    iex> with_sugar = workflow do\n    ...>   stream_source([1, 2, 3])\n    ...>   ~> map(map_fun)\n    ...>   ~> print()\n    ...> end\n    iex> with_sugar == without_sugar\n    true\n\n![](assets/docs/dsl_workflow_bio.png)\n\nThe `as:` and `with:` options (see `node/2`) can be passed as arguments to the operators:\n\n    iex> wf = workflow do\n    ...>   stream_source([1, 2, 3], as: source)\n    ...>   map(fn i -> i * 2 end, as: map)\n    ...>\n    ...>   source._ ~> map._\n    ...> end\n    iex> wf.nodes[:source].operation\n    Skitter.BIO.StreamSource\n    iex> wf.nodes[:map].operation\n    Skitter.BIO.Map\n\n![](assets/docs/dsl_workflow_bio_options.png)\n\nIt is possible to define your own operators and import these into the workflow DSL. For\ninstance:\n\n    iex> wf = workflow do\n    ...>   import MyCustomOperators\n    ...>   my_operator(as: my_operator)\n    ...> end\n    iex> wf.nodes[:my_operator].operation\n    Example\n\n![](assets/docs/dsl_workflow_custom_operator.png)\n\nInformation on how to define custom operators can be found in the\n[Skitter manual](operators.html)","ref":"Skitter.DSL.Workflow.html#workflow/2-nodes-links-and-syntactic-sugar"},{"type":"macro","title":"Examples - Skitter.DSL.Workflow.workflow/2","doc":"iex> workflow in: [foo, bar], out: baz do\n    ...>   foo ~> node(Example) ~> joiner.left\n    ...>   bar ~> node(Example) ~> joiner.right\n    ...>\n    ...>   node(Join, with: SomeStrategy, as: joiner)\n    ...>   ~> node(Example, args: :some_args)\n    ...>   ~> baz\n    ...> end\n    %Skitter.Workflow{\n      in: [\n        foo: [\"skitter/dsl/workflow_test/example#1\": :in_port],\n        bar: [\"skitter/dsl/workflow_test/example#2\": :in_port],\n      ],\n      out: [:baz],\n      nodes: %{\n        \"skitter/dsl/workflow_test/example#1\": %Skitter.Workflow.Node.Operation{\n          operation: Example, args: nil, strategy: DefaultStrategy, links: [out_port: [joiner: :left]]\n        },\n        \"skitter/dsl/workflow_test/example#2\": %Skitter.Workflow.Node.Operation{\n          operation: Example, args: nil, strategy: DefaultStrategy, links: [out_port: [joiner: :right]]\n        },\n        joiner: %Skitter.Workflow.Node.Operation{\n          operation: Join, args: nil, strategy: SomeStrategy, links: [_: [\"skitter/dsl/workflow_test/example#3\": :in_port]]\n        },\n        \"skitter/dsl/workflow_test/example#3\": %Skitter.Workflow.Node.Operation{\n          operation: Example, args: :some_args, strategy: DefaultStrategy, links: [out_port: [:baz]]\n        }\n      }\n    }\n\n![](assets/docs/dsl_workflow_example.png)","ref":"Skitter.DSL.Workflow.html#workflow/2-examples"},{"type":"macro","title":"Skitter.DSL.Workflow.~>/2","doc":"Generate a workflow link.\n\nThis macro connects two ports in the workflow with each other. It can only be used inside\n`workflow/2`.\n\nThis macro can be used in various different ways and provides various conveniences to shorten\nthe definition of a workflow. In its most basic form, this macro is used as follows:\n\n```\nsource ~> destination\n```\n\nWhere source and destination have one of the following two forms:\n\n* ` . `: specifies a port of an operation\n* ` `: specifies a workflow port\n\nFor instance:\n\n    iex> wf = workflow in: foo, out: bar do\n    ...>   node Example, as: node1\n    ...>   node Example, as: node2\n    ...>\n    ...>   foo ~> node1.in_port            # workflow port ~> operation port\n    ...>   node1.out_port ~> node2.in_port # operation port ~> operation port\n    ...>   node2.out_port ~> bar           # operation port ~> workflow port\n    ...> end\n    iex> wf.in\n    [foo: [node1: :in_port]]\n    iex> wf.nodes[:node1].links\n    [out_port: [node2: :in_port]]\n    iex> wf.nodes[:node2].links\n    [out_port: [:bar]]\n\nSome syntactic sugar is present for linking nodes when they are created:\n\n* When the left hand side of `~>` is a node, a link is created between the first out port of\nthis node and the destination.\n* When the right hand side of `~>` is a node, a link is created between the source and the first\nin port of this node.\n\nFor instance:\n\n    iex> wf = workflow in: foo, out: bar do\n    ...>   foo ~> node(Example, as: node1)\n    ...>   node(Example, as: node2) ~> bar\n    ...> end\n    iex> wf.in\n    [foo: [node1: :in_port]]\n    iex> wf.nodes[:node1].links\n    []\n    iex> wf.nodes[:node2].links\n    [out_port: [:bar]]\n\nBoth the left hand side and the right hand side can be nodes:\n\n    iex> wf = workflow do\n    ...>   node(Example, as: node1) ~> node(Example, as: node2)\n    ...> end\n    iex> wf.nodes[:node1].links\n    [out_port: [node2: :in_port]]\n\nFinally, `~>` always returns the right hand side as its result. This enables `~>` to be chained.\n\n    iex> wf = workflow in: foo, out: bar do\n    ...>   foo ~> node(Example, as: node1) ~> node(Example, as: node2) ~> bar\n    ...> end\n    iex> wf.in\n    [foo: [node1: :in_port]]\n    iex> wf.nodes[:node1].links\n    [out_port: [node2: :in_port]]\n    iex> wf.nodes[:node2].links\n    [out_port: [:bar]]","ref":"Skitter.DSL.Workflow.html#~%3E/2"},{"type":"behaviour","title":"Skitter.Operation","doc":"Operation type definition and utilities.\n\nAn operation is a reusable data processing step that can be embedded inside of a workflow. It is\ndefined as the combination of a set of callbacks, which implement the logic of the operation,\nand some metadata, which define how the operation is embedded inside a workflow and how it is\ndistributed over the cluster at runtime.\n\nA skitter operation is defined as an elixir module which implements the `Skitter.Operation`\nbehaviour. This behaviour defines various elixir callbacks which are used to track operation\ninformation such as the defined callbacks. Instead of implementing an operation as an elixir\nmodule, it is recommend to use `Skitter.DSL.Operation.defoperation/3`, which automatically\ngenerates the appropriate callbacks.\n\nThis module defines the operation type and behaviour along with utilities to handle operations.","ref":"Skitter.Operation.html"},{"type":"behaviour","title":"Callbacks - Skitter.Operation","doc":"An operation defines various _callbacks_: functions which implement the data processing logic of\nan operation. These callbacks need to have the ability to modify state and emit data when they\nare called. Callbacks are implemented as elixir functions with a few properties:\n\n- Callbacks accept `t:state/0` and `t:config/0` as their first and second arguments.\n- All other arguments provided to the callback must be wrapped in a `t:Skitter.Token.t/0`.\n- Callbacks return a `t:result/0` struct, which wraps the result of the callback call along with\nthe updated state and emitted data.\n\nBesides this, callbacks track additional information about whether they access or modify state\nand which data they emit. This information is stored inside the behaviour callbacks defined in\nthis module.","ref":"Skitter.Operation.html#module-callbacks"},{"type":"behaviour","title":"Examples - Skitter.Operation","doc":"Since operations need to be defined in a module the example code shown in this module's\ndocumentation assumes the following module is defined:\n\n```\ndefmodule OperationModule do\n  @behaviour Skitter.Operation\n  alias Skitter.Operation.Callback.{Info, Result}\n\n  def _sk_operation_info(:strategy), do: Strategy\n  def _sk_operation_info(:in_ports), do: [:input]\n  def _sk_operation_info(:out_ports), do: [:output]\n\n  def _sk_callbacks, do: MapSet.new(example: 1)\n\n  def _sk_callback_info(:example, 1) do\n    %Info{read?: true, write?: false, emit?: true}\n  end\n\n  def example(state, config, arg) do\n    result = state * config\n    %Result{state: state, emit: [arg: arg], result: result}\n  end\nend\n```","ref":"Skitter.Operation.html#module-examples"},{"type":"callback","title":"Skitter.Operation._sk_callback_info/2","doc":"Return the callback information of callback `name`, `arity`.","ref":"Skitter.Operation.html#c:_sk_callback_info/2"},{"type":"callback","title":"Skitter.Operation._sk_callbacks/0","doc":"Return the names and arities of all the callbacks defined in this module.","ref":"Skitter.Operation.html#c:_sk_callbacks/0"},{"type":"callback","title":"Skitter.Operation._sk_operation_info/1","doc":"Returns the meta-information of the operation.\n\nThe following information is stored:\n\n- `:in_ports`: A list of port names which represents in ports through which the operation\nreceives incoming data.\n\n- `:out_ports`: A list of out ports names which represents the out ports the operation can use\nto emit data.\n\n- `:strategy`: The `Skitter.Strategy` of the operation. `nil` may be provided instead, in which\ncase a strategy must be provided when the operation is embedded in a workflow.","ref":"Skitter.Operation.html#c:_sk_operation_info/1"},{"type":"function","title":"Skitter.Operation.arity/1","doc":"Obtain the arity of `operation`.\n\nThe arity is defined as the amount of in ports the operation defines.","ref":"Skitter.Operation.html#arity/1"},{"type":"function","title":"Examples - Skitter.Operation.arity/1","doc":"iex> arity(OperationModule)\n    1","ref":"Skitter.Operation.html#arity/1-examples"},{"type":"function","title":"Skitter.Operation.call/5","doc":"Call callback `callback_name` with `state`, `config` and `arguments`.\n\nThe provided arguments are automatically wrapped in a `t:Skitter.Token.t/0` using\n`Skitter.Token.wrap/1`.","ref":"Skitter.Operation.html#call/5"},{"type":"function","title":"Examples - Skitter.Operation.call/5","doc":"iex> call(OperationModule, :example, 10, 2, [:foo])\n    %Skitter.Operation.Callback.Result{state: 10, result: 20, emit: [arg: %Token{value: :foo}]}\n    iex> call(OperationModule, :example, 10, 2, [%Token{value: :foo, meta: %{bar: :baz}}])\n    %Skitter.Operation.Callback.Result{state: 10, result: 20, emit: [arg: %Token{value: :foo, meta: %{bar: :baz}}]}","ref":"Skitter.Operation.html#call/5-examples"},{"type":"function","title":"Skitter.Operation.call_if_exists/5","doc":"Call `callback_name` defined by `operation` if it exists.\n\nCalls the callback (using `call/5`) with the given name with `state`, `config` and `args` if\n`{name, length(args)}` exists. If the callback does not exist, a  result with an empty emit\nlist, `nil` as result and `nil` as state is returned.","ref":"Skitter.Operation.html#call_if_exists/5"},{"type":"function","title":"Examples - Skitter.Operation.call_if_exists/5","doc":"iex> call_if_exists(OperationModule, :example, 10, 2, [:foo])\n    %Skitter.Operation.Callback.Result{state: 10, result: 20, emit: [arg: %Token{value: :foo}]}\n    iex> call_if_exists(OperationModule, :example, 10, 2, [:foo, :bar])\n    %Skitter.Operation.Callback.Result{state: nil, result: nil, emit: []}","ref":"Skitter.Operation.html#call_if_exists/5-examples"},{"type":"function","title":"Skitter.Operation.callback_exists?/3","doc":"Check if `operation` defines a callback with `name` and `arity`.","ref":"Skitter.Operation.html#callback_exists?/3"},{"type":"function","title":"Examples - Skitter.Operation.callback_exists?/3","doc":"iex> callback_exists?(OperationModule, :example, 1)\n  true\n\n  iex> callback_exists?(OperationModule, :example, 2)\n  false","ref":"Skitter.Operation.html#callback_exists?/3-examples"},{"type":"function","title":"Skitter.Operation.callback_info/3","doc":"Obtain the callback information for callback `name`, `arity` in `operation`.","ref":"Skitter.Operation.html#callback_info/3"},{"type":"function","title":"Examples - Skitter.Operation.callback_info/3","doc":"iex> callback_info(OperationModule, :example, 1)\n    %Info{read?: true, write?: false, emit?: true}","ref":"Skitter.Operation.html#callback_info/3-examples"},{"type":"function","title":"Skitter.Operation.callbacks/1","doc":"Obtain the set of all callbacks defined in `operation`.","ref":"Skitter.Operation.html#callbacks/1"},{"type":"function","title":"Examples - Skitter.Operation.callbacks/1","doc":"iex> callbacks(OperationModule)\nMapSet.new([example: 1])","ref":"Skitter.Operation.html#callbacks/1-examples"},{"type":"function","title":"Skitter.Operation.in_port_to_index/2","doc":"Get the index of an in port.","ref":"Skitter.Operation.html#in_port_to_index/2"},{"type":"function","title":"Examples - Skitter.Operation.in_port_to_index/2","doc":"iex> in_port_to_index(OperationModule, :input)\n    0\n    iex> in_port_to_index(OperationModule, :other)\n    nil","ref":"Skitter.Operation.html#in_port_to_index/2-examples"},{"type":"function","title":"Skitter.Operation.in_ports/1","doc":"Obtain the list of in ports of `operation`.","ref":"Skitter.Operation.html#in_ports/1"},{"type":"function","title":"Examples - Skitter.Operation.in_ports/1","doc":"iex> in_ports(OperationModule)\n    [:input]","ref":"Skitter.Operation.html#in_ports/1-examples"},{"type":"function","title":"Skitter.Operation.index_to_in_port/2","doc":"Convert an index of an in port to a port name.","ref":"Skitter.Operation.html#index_to_in_port/2"},{"type":"function","title":"Examples - Skitter.Operation.index_to_in_port/2","doc":"iex> index_to_in_port(OperationModule, 0)\n    :input\n    iex> index_to_in_port(OperationModule, 1)\n    nil","ref":"Skitter.Operation.html#index_to_in_port/2-examples"},{"type":"function","title":"Skitter.Operation.index_to_out_port/2","doc":"Convert an index of an out port to a port name.","ref":"Skitter.Operation.html#index_to_out_port/2"},{"type":"function","title":"Examples - Skitter.Operation.index_to_out_port/2","doc":"iex> index_to_out_port(OperationModule, 0)\n    :output\n    iex> index_to_out_port(OperationModule, 1)\n    nil","ref":"Skitter.Operation.html#index_to_out_port/2-examples"},{"type":"function","title":"Skitter.Operation.initial_state/2","doc":"Obtain the initial state for `operation`.\n\nSome operations handle stateful logic. Often, these operations have the notion of an \"initial\nstate\": the state the operation has before it received any data records.\n\nBy convention, this initial state is defined as a callback named `initial_state`. This function\ncalls this callback if it exists with an empty argument list and the provided configuration (or\n`nil`, if not configuration is passed). If the `initial_state` callback is not defined, `nil` is\nreturned.","ref":"Skitter.Operation.html#initial_state/2"},{"type":"function","title":"Examples - Skitter.Operation.initial_state/2","doc":"iex> initial_state(OperationModule)\n    nil\n\nIf `InitialStateModule` is defined as follows:\n\n```\ndefmodule InitialStateModule do\n  @behaviour Skitter.Operation\n  alias Skitter.Operation.Callback.{Info, Result}\n\n  def _sk_operation_info(:strategy), do: Strategy\n  def _sk_operation_info(:in_ports), do: []\n  def _sk_operation_info(:out_ports), do: []\n\n  def _sk_callbacks, do: MapSet.new(initial_state: 0)\n\n  def _sk_callback_info(:initial_state, 0) do\n    %Info{read?: false, write?: false, emit?: false}\n  end\n\n  def initial_state(_, :foo), do: %Result{result: :bar}\n  def initial_state(_, _), do: %Result{result: 42}\nend\n```\n\n    iex> initial_state(InitialStateModule)\n    42\n    iex> initial_state(InitialStateModule, :foo)\n    :bar","ref":"Skitter.Operation.html#initial_state/2-examples"},{"type":"function","title":"Skitter.Operation.operation?/1","doc":"Check if a given value refers to an operation module.","ref":"Skitter.Operation.html#operation?/1"},{"type":"function","title":"Examples - Skitter.Operation.operation?/1","doc":"iex> operation?(5)\n    false\n    iex> operation?(String)\n    false\n    iex> operation?(OperationModule)\n    true","ref":"Skitter.Operation.html#operation?/1-examples"},{"type":"function","title":"Skitter.Operation.out_port_to_index/2","doc":"Get the index of an out port.","ref":"Skitter.Operation.html#out_port_to_index/2"},{"type":"function","title":"Examples - Skitter.Operation.out_port_to_index/2","doc":"iex> out_port_to_index(OperationModule, :output)\n    0\n    iex> out_port_to_index(OperationModule, :other)\n    nil","ref":"Skitter.Operation.html#out_port_to_index/2-examples"},{"type":"function","title":"Skitter.Operation.out_ports/1","doc":"Obtain the list of out ports of `operation`.","ref":"Skitter.Operation.html#out_ports/1"},{"type":"function","title":"Examples - Skitter.Operation.out_ports/1","doc":"iex> out_ports(OperationModule)\n    [:output]","ref":"Skitter.Operation.html#out_ports/1-examples"},{"type":"function","title":"Skitter.Operation.sink?/1","doc":"Check if an operation is a sink.\n\nAn operation is a sink if it does not have any out ports.","ref":"Skitter.Operation.html#sink?/1"},{"type":"function","title":"Examples - Skitter.Operation.sink?/1","doc":"iex> sink?(OperationModule)\n    false","ref":"Skitter.Operation.html#sink?/1-examples"},{"type":"function","title":"Skitter.Operation.source?/1","doc":"Check if an operation is a source.\n\nAn operation is a source if it does not have any in ports.","ref":"Skitter.Operation.html#source?/1"},{"type":"function","title":"Examples - Skitter.Operation.source?/1","doc":"iex> source?(OperationModule)\n    false","ref":"Skitter.Operation.html#source?/1-examples"},{"type":"function","title":"Skitter.Operation.strategy/1","doc":"Obtain the default strategy of `operation`.","ref":"Skitter.Operation.html#strategy/1"},{"type":"function","title":"Examples - Skitter.Operation.strategy/1","doc":"iex> strategy(OperationModule)\n    Strategy","ref":"Skitter.Operation.html#strategy/1-examples"},{"type":"type","title":"Skitter.Operation.args/0","doc":"Arguments passed to a callback when it is called.\n\nThe arguments are provided as tokens or plain values wrapped in a list.","ref":"Skitter.Operation.html#t:args/0"},{"type":"type","title":"Skitter.Operation.config/0","doc":"Configuration passed to the callback when it is called.\n\nThe configuration represents the immutable state of an operation. It is explicitly separated\nfrom the mutable `t:state/0` to enable strategies to explicitly differentiate between handling\nmutable and immutable data.","ref":"Skitter.Operation.html#t:config/0"},{"type":"type","title":"Skitter.Operation.emit/0","doc":"Output emitted by a callback.\n\nEmitted data is returned as a keyword list where the output for each out port is specified. When\nno data is emitted on a port, the port should be omitted from the list. The data emitted by a\ncallback for a port should be wrapped in an `t:Enumerable.t/0`. Each element in this enumerable\nwill be sent to downstream nodes separately.\n\nThe data emitted for a port (i.e. the data inside the enumerable) may either be a \"plain\" Elixir\nvalue or a Skitter token. The Skitter runtime will automatically wrap plain values in a Skitter\ntoken when they are sent to downstream operations.","ref":"Skitter.Operation.html#t:emit/0"},{"type":"type","title":"Skitter.Operation.info/0","doc":"Additional callback information. Can be retrieved with `info/2`.\n\nThe following information is stored:\n\n- `:read?`: Boolean which indicates if the callback reads the operation state.\n- `:write?`: Boolean which indicates if the callback updates the operation state.\n- `:emit`: Boolean which indicates if the callback emits data.","ref":"Skitter.Operation.html#t:info/0"},{"type":"type","title":"Skitter.Operation.port_index/0","doc":"Input/output interface of Skitter operations.\n\nThe ports of an operation determine its external interface. A port can be referred to by its\nname, which is stored as an atom. Inside strategies, it may be useful to use the index of a port\nto remain agnostic to its name. `in_port_to_index/2` and `out_port_to_index/2` can be used to\nconvert a port to its index.","ref":"Skitter.Operation.html#t:port_index/0"},{"type":"type","title":"Skitter.Operation.port_name/0","doc":"Input/output interface of Skitter operations.\n\nThe ports of an operation determine its external interface. A port can be referred to by its\nname, which is stored as an atom. Inside strategies, it may be useful to use the index of a port\nto remain agnostic to its name. `in_port_to_index/2` and `out_port_to_index/2` can be used to\nconvert a port to its index.","ref":"Skitter.Operation.html#t:port_name/0"},{"type":"type","title":"Skitter.Operation.result/0","doc":"Values returned by a callback when it is called.\n\nThe following information is stored:\n\n- `:result`: The actual result of the callback, i.e. the final value returned in its body.\n- `:state`: The (possibly modified) state after calling the callback.\n- `:emit`: The output emitted by the callback.","ref":"Skitter.Operation.html#t:result/0"},{"type":"type","title":"Skitter.Operation.state/0","doc":"State passed to the callback when it is called.\n\nThe state represents the mutable state of an operation. It is explicitly separated from the\nimmutable `t:config/0` to enable strategies to explicitly differentiate between handling mutable\nand immutable data.","ref":"Skitter.Operation.html#t:state/0"},{"type":"type","title":"Skitter.Operation.t/0","doc":"An operation is defined as a module.\n\nThis module should implement the `Skitter.Operation` behaviour.","ref":"Skitter.Operation.html#t:t/0"},{"type":"module","title":"Skitter.Strategy","doc":"Strategy type definition and utilities.\n\nA strategy is a reusable piece of logic which determines how an operation is distributed at\nruntime. It is defined as a collection of _hooks_: functions which each define an aspect of the\ndistributed behaviour of an operation.\n\nAn operation strategy is defined as an elixir module which implements the\n`Skitter.Strategy.Operation` behaviour. It is recommended to define a strategy using\n`Skitter.DSL.Strategy.defstrategy/3`.\n\nThis module defines the strategy and context types.","ref":"Skitter.Strategy.html"},{"type":"type","title":"Skitter.Strategy.context/0","doc":"Context information for strategy hooks.\n\nA strategy hook often needs information about the context in which it is being called. Relevant\ninformation about the context is stored inside the context, which is passed as the first\nargument to every hook.\n\nThe following information is stored:\n\n- `operation`: The operation for which the hook is called.\n- `strategy`: The strategy of the operation.\n- `deployment`: The current deployment data. `nil` if the deployment is not created yet (e.g. in\n`deploy`)\n- `_skr`: Data stored by the runtime system. This data should not be accessed or modified.","ref":"Skitter.Strategy.html#t:context/0"},{"type":"type","title":"Skitter.Strategy.deployment/0","doc":"Immutable data of a data processing pipeline.\n\nA strategy which is deployed over the cluster has access to an immutable set of data which is\ntermed the _deployment_. A strategy can specify which data to store in its deployment inside the\n`c:Skitter.Strategy.Operation.deploy/2` hook. Afterwards, the other strategy hooks have access\nto the data stored within the deployment.\n\nNote that an operation strategy can only access its own deployment data.","ref":"Skitter.Strategy.html#t:deployment/0"},{"type":"type","title":"Skitter.Strategy.t/0","doc":"A strategy is defined as a module.","ref":"Skitter.Strategy.html#t:t/0"},{"type":"module","title":"Skitter.Token","doc":"Metadata tracking wrapper\n\nA data element being processed by a Skitter workflow is wrapped inside a _token_. This token is\nused to track additional information about the data. The Skitter runtime system stores the port\non which the data record was received inside the token, while operations and strategies are free\nto add arbitrary meta-information.","ref":"Skitter.Token.html"},{"type":"module","title":"Automatic (un)wrapping - Skitter.Token","doc":"When an operation callback is called through the use of `Skitter.Operation.call/5` or\n`Skitter.Operation.call_if_exists/5`, `wrap/1` will be used to wrap any non-token argument\ninside a token. Thus, _every_ argument accepted by a callback is always wrapped inside a token.\n\nThe operation DSL (i.e. `Skitter.DSL.Operation.defcb/2`) automatically extracts the values from\nthese tokens, and enables the meta-information contained in the tokens to be accessed through\nthe use of `Skitter.DSL.Operation.port_of/1` and `Skitter.DSL.Operation.meta_of/1`.\n\nAn operation may emit plain values or tokens, which are then received by the strategy. Any data\nemitted by the strategy (through the use of `Skitter.Strategy.Operation.emit/2` or\n`Skitter.DSL.Strategy.Helpers.emit/1`) is automatically wrapped inside a token.","ref":"Skitter.Token.html#module-automatic-un-wrapping"},{"type":"module","title":"Available Information - Skitter.Token","doc":"","ref":"Skitter.Token.html#module-available-information"},{"type":"module","title":"Port - Skitter.Token","doc":"When a data record is provided to a strategy through `c:Skitter.Strategy.Operation.deliver/2`,\nthe Skitter runtime system ensures that the data record is wrapped inside a token which stores\nthe `port` on which the token was received. If this token is passed to an operation callback\nunmodified, `Skitter.DSL.Operation.port_of/1` can be used to obtain this port.","ref":"Skitter.Token.html#module-port"},{"type":"module","title":"Meta-information - Skitter.Token","doc":"An operation or strategy may add arbitrary meta-information to a token. This information can be\nused to track application-specific concerns such as timing information (used to implement\nwindowing support). The Skitter runtime system does not handle this information in any way.\nIt is therefore up to the developers of operations and strategies to ensure that this\ninformation is not lost.","ref":"Skitter.Token.html#module-meta-information"},{"type":"function","title":"Skitter.Token.put_meta/3","doc":"Add meta-information to a token or value.\n\nThis function stores a key-value pair in the meta-information of a token. If a plain value is\nprovided, it is wrapped inside a token first. If a value was present for `key` in the\nmeta-information of token, it is overwritten.","ref":"Skitter.Token.html#put_meta/3"},{"type":"function","title":"Examples - Skitter.Token.put_meta/3","doc":"iex> put_meta(%Token{value: 5, port: :foo, meta: %{window: 10}}, :foo, :bar)\n    %Token{value: 5, port: :foo, meta: %{window: 10, foo: :bar}}\n    iex> put_meta(%Token{value: 5, port: :foo, meta: %{window: 10}}, :window, 20)\n    %Token{value: 5, port: :foo, meta: %{window: 20}}\n    iex> put_meta(5, :foo, :bar)\n    %Token{value: 5, port: nil, meta: %{foo: :bar}}","ref":"Skitter.Token.html#put_meta/3-examples"},{"type":"function","title":"Skitter.Token.unwrap/1","doc":"Extract the value from a token.\n\nGet the value from a token, or return a value if no token is provided.","ref":"Skitter.Token.html#unwrap/1"},{"type":"function","title":"Examples - Skitter.Token.unwrap/1","doc":"iex> unwrap(%Token{value: 5})\n    5\n    iex> unwrap(5)\n    5","ref":"Skitter.Token.html#unwrap/1-examples"},{"type":"function","title":"Skitter.Token.update_meta/3","doc":"Update meta-information associated with a value.\n\nThis function obtains the value for `key` stored inside the meta-information of a token and\npasses it to the provided function. The result of the function is stored as the new value for\n`key` in the meta-information of the token.\n\nIf a plain value is provided, the value is wrapped inside a token first. In this case, or if\n`key` does not exist in the meta-information of the provided token, the function is called with\nvalue `nil`.","ref":"Skitter.Token.html#update_meta/3"},{"type":"function","title":"Examples - Skitter.Token.update_meta/3","doc":"iex> update_meta(%Token{value: 5, port: :foo, meta: %{window: 10}}, :window, fn\n    ...>   nil -> 0\n    ...>   window -> window + 10\n    ...> end)\n    %Token{value: 5, port: :foo, meta: %{window: 20}}\n    iex> update_meta(%Token{value: 5, port: :foo, meta: %{foo: :bar}}, :window, fn\n    ...>   nil -> 0\n    ...>   window -> window + 10\n    ...> end)\n    %Token{value: 5, port: :foo, meta: %{window: 0, foo: :bar}}\n    iex> update_meta(5, :window, fn\n    ...>   nil -> 0\n    ...>   window -> window + 10\n    ...> end)\n    %Token{value: 5, port: nil, meta: %{window: 0}}","ref":"Skitter.Token.html#update_meta/3-examples"},{"type":"function","title":"Skitter.Token.with_meta/2","doc":"Set the meta-information of a token or value.\n\nThis function associates meta-information with a token. If a plain value is provided, it is\nwrapped in a token first. If a token with existing meta-information is provided, the existing\nmeta-information is overwritten.","ref":"Skitter.Token.html#with_meta/2"},{"type":"function","title":"Examples - Skitter.Token.with_meta/2","doc":"iex> with_meta(%Token{value: 5, port: :foo, meta: %{window: 10}}, foo: :bar, window: 20)\n    %Token{value: 5, port: :foo, meta: %{window: 20, foo: :bar}}\n    iex> with_meta(5, foo: :bar, window: 20)\n    %Token{value: 5, port: nil, meta: %{window: 20, foo: :bar}}","ref":"Skitter.Token.html#with_meta/2-examples"},{"type":"function","title":"Skitter.Token.wrap/1","doc":"Wrap a value in a token.\n\nThis operation is a no-op if it receives a token. Otherwise, it wraps the value inside a token\nwith no port and empty meta-information.","ref":"Skitter.Token.html#wrap/1"},{"type":"function","title":"Examples - Skitter.Token.wrap/1","doc":"iex> wrap(%Token{value: 5, port: :foo, meta: %{window: 10}})\n    %Token{value: 5, port: :foo, meta: %{window: 10}}\n    iex> wrap(5)\n    %Token{value: 5, port: nil, meta: %{}}","ref":"Skitter.Token.html#wrap/1-examples"},{"type":"type","title":"Skitter.Token.t/0","doc":"Values tracked by a token.\n\nA token tracks the value it wraps (`:value`), the port on which the value was received\n(`:port`), and a collection of arbitrary meta-information (`:meta`). The meta-information is\nstored as key-value pairs.","ref":"Skitter.Token.html#t:t/0"},{"type":"module","title":"Skitter.Workflow","doc":"Workflow type definition and utilities.\n\nA workflow defines a data processing pipeline. It is defined as a set of nodes connected through\nvarious links. A node contains a data processing operation and its distribution strategy. A\nworkflow stores these nodes and links between them, along with additional meta-information about\nthe workflow.\n\nIn order to enable the reuse of workflows, workflows may define in -and out ports. When this is\ndone, these workflows may be embedded inside another workflow. Note that a workflow is always\nflattened using `flatten/1` before it is deployed.\n\nThis module defines the workflow type along with some utilities to work with this type. It is\nnot recommended to define a workflow manually. Instead, the use of\n`Skitter.DSL.Workflow.workflow/2` is preferred.","ref":"Skitter.Workflow.html"},{"type":"function","title":"Skitter.Workflow.flatten/1","doc":"Recursively inline any nested workflow of a workflow.\n\nThis function ensures any workflow embedded in the provided workflow is inlined into the\nprovided workflow.","ref":"Skitter.Workflow.html#flatten/1"},{"type":"function","title":"Examples - Skitter.Workflow.flatten/1","doc":"![](assets/docs/workflow_inline_before.png)\nwill be converted to:\n![](assets/docs/workflow_inline_after.png)\n\n    iex> defoperation Simple, in: p, out: p do\n    ...> end\n    iex> defoperation Join, in: [left, right], out: p do\n    ...> end\n    iex> inner = %Workflow{\n    ...>   in: [foo: [node1: :p, node2: :p]],\n    ...>   out: [:bar],\n    ...>   nodes: %{\n    ...>     node1: %Node.Operation{operation: Simple, links: [p: [node3: :left]]},\n    ...>     node2: %Node.Operation{operation: Simple, links: [p: [node3: :right]]},\n    ...>     node3: %Node.Operation{operation: Join, links: [p: [:bar]]}\n    ...> }}\n    iex> parent = %Workflow{\n    ...>   nodes: %{\n    ...>     node_pre: %Node.Operation{operation: Simple, links: [p: [nested1: :foo, nested2: :foo]]},\n    ...>     nested1: %Node.Workflow{workflow: inner, links: [bar: [node_post: :left]]},\n    ...>     nested2: %Node.Workflow{workflow: inner, links: [bar: [node_post: :right]]},\n    ...>     node_post: %Node.Operation{operation: Join}\n    ...> }}\n    iex> flatten(parent)\n    %Workflow{\n      nodes: %{\n        node_pre: %Node.Operation{operation: Skitter.WorkflowTest.Simple, links: [p: [\"nested1#node1\": :p, \"nested1#node2\": :p, \"nested2#node1\": :p, \"nested2#node2\": :p]]},\n        \"nested1#node1\": %Node.Operation{operation: Skitter.WorkflowTest.Simple, links: [p: [\"nested1#node3\": :left]]},\n        \"nested1#node2\": %Node.Operation{operation: Skitter.WorkflowTest.Simple, links: [p: [\"nested1#node3\": :right]]},\n        \"nested1#node3\": %Node.Operation{operation: Skitter.WorkflowTest.Join, links: [p: [node_post: :left]]},\n        \"nested2#node1\": %Node.Operation{operation: Skitter.WorkflowTest.Simple, links: [p: [\"nested2#node3\": :left]]},\n        \"nested2#node2\": %Node.Operation{operation: Skitter.WorkflowTest.Simple, links: [p: [\"nested2#node3\": :right]]},\n        \"nested2#node3\": %Node.Operation{operation: Skitter.WorkflowTest.Join, links: [p: [node_post: :right]]},\n        node_post: %Node.Operation{operation: Skitter.WorkflowTest.Join}\n      }\n    }","ref":"Skitter.Workflow.html#flatten/1-examples"},{"type":"function","title":"Skitter.Workflow.verify/1","doc":"Verify if the links in a workflow are valid.\n\nThis function verifies if every link in the workflow has a valid source and destination. That\nis, the link should depart from an existing workflow or operation port and arrive at one. Note\nthat this function does _not_ traverse nested workflows.","ref":"Skitter.Workflow.html#verify/1"},{"type":"function","title":"Examples - Skitter.Workflow.verify/1","doc":"iex> defoperation Example, in: p, out: p do\n    ...> end\n    iex> verify(%Workflow{nodes: %{\n    ...>   foo: %Node.Operation{operation: Example, links: [p: [bar: :p]]},\n    ...>   bar: %Node.Operation{operation: Example},\n    ...> }})\n    :ok\n    iex> verify(%Workflow{nodes: %{\n    ...>   foo: %Node.Operation{operation: Example, links: [p: [baz: :p]]},\n    ...>   bar: %Node.Operation{operation: Example},\n    ...> }})\n    [{{:foo, :p}, {:baz, :p}}]","ref":"Skitter.Workflow.html#verify/1-examples"},{"type":"function","title":"Skitter.Workflow.verify!/1","doc":"Verify if the links in a workflow are valid using `verify/1`.\n\nThis function uses `verify/1` to verify if every link in a workflow has a valid source and\ndestination. If this is not the case, it raises a `Skitter.DefinitionError`. When the workflow\nis valid, the worfklow itself is returned.","ref":"Skitter.Workflow.html#verify!/1"},{"type":"function","title":"Examples - Skitter.Workflow.verify!/1","doc":"iex> defoperation Example, in: p, out: p do\n    ...> end\n    iex> verify!(%Workflow{nodes: %{\n    ...>   foo: %Node.Operation{operation: Example, links: [p: [bar: :p]]},\n    ...>   bar: %Node.Operation{operation: Example},\n    ...> }})\n    %Workflow{nodes: %{\n      foo: %Node.Operation{operation: Skitter.WorkflowTest.Example, links: [p: [bar: :p]]},\n      bar: %Node.Operation{operation: Skitter.WorkflowTest.Example},\n    }}\n    iex> verify!(%Workflow{nodes: %{\n    ...>   foo: %Node.Operation{operation: Example, links: [p: [baz: :p]]},\n    ...>   bar: %Node.Operation{operation: Example},\n    ...> }})\n    ** (Skitter.DefinitionError) Invalid link: {:foo, :p} ~> {:baz, :p}","ref":"Skitter.Workflow.html#verify!/1-examples"},{"type":"type","title":"Skitter.Workflow.args/0","doc":"Operation initialization arguments\n\nThis type stores the arguments passed to the operation in the workflow definition.","ref":"Skitter.Workflow.html#t:args/0"},{"type":"type","title":"Skitter.Workflow.destination/0","doc":"Link destination.\n\nThis type stores the destination of a link. A link can point to a port of a node or to an out\nport of the workflow. In the first case, the name of the node and its out port are stored, in\nthe second, only the name of the out port is stored.","ref":"Skitter.Workflow.html#t:destination/0"},{"type":"type","title":"Skitter.Workflow.links/0","doc":"Collection of outgoing links.\n\nLinks are stored as a keyword list. Each key in this list represents an out port, while the\nvalue of this key is a list which references the destinations of this out port.","ref":"Skitter.Workflow.html#t:links/0"},{"type":"type","title":"Skitter.Workflow.name/0","doc":"Instance name\n\nA name is used to refer to a node embedded inside a workflow.","ref":"Skitter.Workflow.html#t:name/0"},{"type":"type","title":"Skitter.Workflow.operation_node/0","doc":"Operation embedded inside a workflow.\n\nAn operation in a workflow is stored along with its strategy, initialization arguments (which\nare passed to `c:Skitter.Strategy.Operation.deploy/2`) and the outgoing links of each of its out\nports.\n\nWorkflows can override the strategy of an operation, therefore, the strategy specified here may\nnot be the same as the strategy returned by `Skitter.Operation.strategy/1`.","ref":"Skitter.Workflow.html#t:operation_node/0"},{"type":"type","title":"Skitter.Workflow.t/0","doc":"Internal workflow representation.\n\nA workflow is stored as a map, where each name refers to a single node, which is either a\n`t:operation/0` or `t:workflow/0`. Besides this, the in -and out ports of the workflow are\nstored. The outgoing links of the in ports of a workflow are stored along with the in ports.","ref":"Skitter.Workflow.html#t:t/0"},{"type":"type","title":"Skitter.Workflow.workflow_node/0","doc":"Workflow embedded inside a workflow.\n\nA workflow nested inside a workflow is stored along with the outgoing links of its out ports.","ref":"Skitter.Workflow.html#t:workflow_node/0"},{"type":"behaviour","title":"Skitter.Strategy.Operation","doc":"Operation strategy behaviour.\n\nThis module defines and documents the various hooks a `Skitter.Strategy` for an operation should\nimplement, along with the functions it can use to access the runtime system.","ref":"Skitter.Strategy.Operation.html"},{"type":"callback","title":"Skitter.Strategy.Operation.deliver/2","doc":"Accept data sent to the operation node and send it to a worker.\n\nThis hook is called by the runtime system when data needs to be sent to a given operation (i.e.\nwhen a predecessor of the operation node emits data). It receives the data to be sent wrapped in\na `t:Skitter.Token.t/0`, which contains the name of the port to which the data should be sent.\n\nThe result of this hook is ignored. Instead, this hook should use `Skitter.Worker.send/2` to\ntransfer the received data to a worker.","ref":"Skitter.Strategy.Operation.html#c:deliver/2"},{"type":"callback","title":"Context - Skitter.Strategy.Operation.deliver/2","doc":"All context data (operation, strategy and deployment data) is available when this hook is\ncalled.","ref":"Skitter.Strategy.Operation.html#c:deliver/2-context"},{"type":"callback","title":"Skitter.Strategy.Operation.deploy/2","doc":"Deploy an operation over the cluster.\n\nThis hook is called by the runtime system when an operation has to be distributed over the\ncluster. It receives the arguments passed to the operation's node in the workflow definition.\nAny data returned by this hook is made available to other hooks through the `deployment` field\nin `t:Skitter.Strategy.context/0`.","ref":"Skitter.Strategy.Operation.html#c:deploy/2"},{"type":"callback","title":"Context - Skitter.Strategy.Operation.deploy/2","doc":"When this hook is called, only the current strategy and operation are available in the context.","ref":"Skitter.Strategy.Operation.html#c:deploy/2-context"},{"type":"function","title":"Skitter.Strategy.Operation.emit/2","doc":"Emit values.\n\nThis function causes the current operation node to emit data. In other words, the provided data\nwill be sent to the operation nodes connected to the out ports of the current operation node.\nThis function accepts a keyword list of `{out_port, enum}` pairs. Each element in `enum` will be\nsent to the in ports of the operation nodes connected to `out_port`.\n\nNote that data is emitted from the current worker. This may cause issues when infinite streams\nof data are emitted.","ref":"Skitter.Strategy.Operation.html#emit/2"},{"type":"callback","title":"Skitter.Strategy.Operation.process/4","doc":"Handle a message received by a worker.\n\nThis hook is called by the runtime when a worker process receives a message. It is called with\nthe received message, the data of the worker that received the message and its role. This hook\nshould return the new state of the worker that received the message.","ref":"Skitter.Strategy.Operation.html#c:process/4"},{"type":"callback","title":"Context - Skitter.Strategy.Operation.process/4","doc":"All context data (operation, strategy and the deployment data) is available when this hook is\ncalled.","ref":"Skitter.Strategy.Operation.html#c:process/4-context"},{"type":"module","title":"Skitter.Remote","doc":"Facilities to interact with remote Skitter runtimes.\n\nThis module offers facilities to query Skitter about the available remote Skitter runtimes and\ntheir properties. It also defines various functions which can be used to spawn Skitter workers\non remote Skitter nodes.","ref":"Skitter.Remote.html"},{"type":"function","title":"Skitter.Remote.connect/2","doc":"Attempt to connect to `remote`.\n\nThis function verifies if `remote` is a valid Skitter node and tries to connect to it. If the\nconnection is successful, this functions tries to find a bound handler for the local mode. If\nthis works, `{:ok, mode}` is returned, where `mode` is the mode of the remote runtime.\n\nIf the `expected_mode` argument is passed, `connect/2` only connects to the remote runtime if\nits mode is equal to `expected_mode`.\n\nThe following errors may be returned if the connection does not succeed:\n- `{:error, :not_distributed}`: the local node is not alive according to `Node.alive?/0`\n- `{:error, :not_connected}`: it was not possible to connect to the remote runtime\n- `{:error, :not_skitter}`: the remote node is not a Skitter remote\n- `{:error, :incompatible}`: the remote node is running an incompatible version of Skitter\n- `{:error, :mode_mismatch}`: the mode of the remote runtime is not equal to `expected_mode`\n- `{:error, :unknown_mode}`: the remote or local runtime does not have a handler for the mode of\nthe other runtime.\n- The handler of the local mode may return a custom `{:error, reason}` tuple.","ref":"Skitter.Remote.html#connect/2"},{"type":"function","title":"Skitter.Remote.connected?/1","doc":"Check if the local runtime is connected to the specified remote runtime.","ref":"Skitter.Remote.html#connected?/1"},{"type":"function","title":"Skitter.Remote.master/0","doc":"Get the name of the master node of the cluster.","ref":"Skitter.Remote.html#master/0"},{"type":"function","title":"Skitter.Remote.n_times_on_all_workers/2","doc":"Execute a function n times on every worker runtime.\n\nA list of results will be returned for each worker node. These results will be returned in a\nkeyword list of `{worker, result}` pairs.","ref":"Skitter.Remote.html#n_times_on_all_workers/2"},{"type":"function","title":"Skitter.Remote.on/2","doc":"Execute `fun` on `remote`, block until a result is available.","ref":"Skitter.Remote.html#on/2"},{"type":"function","title":"Skitter.Remote.on/4","doc":"Execute `mod.func(args)` on `remote`, block until a result is available.","ref":"Skitter.Remote.html#on/4"},{"type":"function","title":"Skitter.Remote.on_all_worker_cores/1","doc":"Execute a function on every core on every worker runtime.\n\nA list of results will be returned for each worker node. These results will be returned in a\nkeyword list of `{worker, result}` pairs.","ref":"Skitter.Remote.html#on_all_worker_cores/1"},{"type":"function","title":"Skitter.Remote.on_all_workers/1","doc":"Execute a function on every worker runtime.\n\nThe result of each worker will be returned in a keyword list of `{worker, result}` pairs.","ref":"Skitter.Remote.html#on_all_workers/1"},{"type":"function","title":"Skitter.Remote.on_all_workers/3","doc":"Execute a function on every worker runtime.\n\nThe result of each worker will be returned in a keyword list of `{worker, result}` pairs.","ref":"Skitter.Remote.html#on_all_workers/3"},{"type":"function","title":"Skitter.Remote.on_many/2","doc":"Execute `fun` on every specified remote runtime, obtain results in a list.","ref":"Skitter.Remote.html#on_many/2"},{"type":"function","title":"Skitter.Remote.on_many/4","doc":"Execute `mod.func(args)` on every specified remote runtime, obtain results in a list.","ref":"Skitter.Remote.html#on_many/4"},{"type":"function","title":"Skitter.Remote.on_n/2","doc":"Execute a function n times, distributed over the available workers.\n\nThis is handy when you wish to create n workers distributed over the cluster. The work to be\ndone will be divided over the worker nodes in a round robin fashion. This behaviour may change\nin the future.","ref":"Skitter.Remote.html#on_n/2"},{"type":"function","title":"Skitter.Remote.on_tagged_worker_cores/2","doc":"Execute a function on every core on every worker runtime tagged with `tag`.\n\nThe result of each worker will be returned in a keyword list of `{worker, result}` pairs.","ref":"Skitter.Remote.html#on_tagged_worker_cores/2"},{"type":"function","title":"Skitter.Remote.on_tagged_workers/2","doc":"Execute a function on every worker runtime tagged with `tag`.\n\nThe result of each worker will be returned in a keyword list of `{worker, result}` pairs.","ref":"Skitter.Remote.html#on_tagged_workers/2"},{"type":"function","title":"Skitter.Remote.self/0","doc":"Get the name of the current Skitter runtime.","ref":"Skitter.Remote.html#self/0"},{"type":"function","title":"Skitter.Remote.tags/1","doc":"Get a list of all the tags of `node()`.","ref":"Skitter.Remote.html#tags/1"},{"type":"function","title":"Skitter.Remote.with_tag/1","doc":"Get a list of all the worker runtimes tagged with a given `t:tag/0`.","ref":"Skitter.Remote.html#with_tag/1"},{"type":"function","title":"Skitter.Remote.workers/0","doc":"Get a list of the names of all the worker runtimes in the cluster.","ref":"Skitter.Remote.html#workers/0"},{"type":"type","title":"Skitter.Remote.tag/0","doc":"Worker tag.\n\nA worker may be started with a tag, which indicates properties of the node. For instance, a\n`:gpu` tag could be added to a node which has a gpu. Various functions in this module can be\nused to only spawn workers on nodes with given tags.","ref":"Skitter.Remote.html#t:tag/0"},{"type":"module","title":"Skitter.Runtime","doc":"Skitter runtime system API.\n\nThis modules defines various functions which can be used to query the Skitter runtime system.\nMany of these functions are only useful to obtain additional information based on data returned\nby [telemetry events](telemetry.md).","ref":"Skitter.Runtime.html"},{"type":"function","title":"Skitter.Runtime.deploy/1","doc":"Deploy a workflow over the cluster.\n\nStarts a Skitter application (i.e. a `t:Skitter.Workflow.t/0`) by deploying it over the cluster.\nThe workflow is flattened (using `Skitter.Workflow.flatten/1`) before it is deployed. After\ndeployment, this function returns a `t:ref/0`, which can be used by various functions in this\nmodule.","ref":"Skitter.Runtime.html#deploy/1"},{"type":"function","title":"Skitter.Runtime.get_workflow/1","doc":"Get the workflow for a reference or context.\n\nThe workflow will not have any in or out ports and only contain operation nodes, since\n`deploy/1` flattens workflows before it deploys them.","ref":"Skitter.Runtime.html#get_workflow/1"},{"type":"function","title":"Skitter.Runtime.mode/0","doc":"Get the current runtime mode.\n\nThis function returns the mode of the current runtime. The available modes and their goal are\ndocumented in the [configuration documentation](configuration.html#modes). This function may\nalso return `:test`, which is only used for testing.","ref":"Skitter.Runtime.html#mode/0"},{"type":"function","title":"Skitter.Runtime.node_for_context/1","doc":"Get the workflow node based on a context.\n\nThe workflow node will always be an operation node.","ref":"Skitter.Runtime.html#node_for_context/1"},{"type":"function","title":"Skitter.Runtime.node_name_for_context/1","doc":"Get the name of the workflow node based on a context.","ref":"Skitter.Runtime.html#node_name_for_context/1"},{"type":"function","title":"Skitter.Runtime.ref_for_context/1","doc":"Get the reference based on a context.\n\nThis can be used to link a telemetry event which contained a context to a deployed workflow.","ref":"Skitter.Runtime.html#ref_for_context/1"},{"type":"function","title":"Skitter.Runtime.spawned_workflows/0","doc":"Get a list with references to every spawned workflow.\n\nThis function communicates with the master node when called from a worker runtime.","ref":"Skitter.Runtime.html#spawned_workflows/0"},{"type":"function","title":"Skitter.Runtime.stop/1","doc":"Stop the workflow with reference `ref`.","ref":"Skitter.Runtime.html#stop/1"},{"type":"type","title":"Skitter.Runtime.ref/0","doc":"Reference to a deployed workflow.","ref":"Skitter.Runtime.html#t:ref/0"},{"type":"module","title":"Skitter.Worker","doc":"Worker which manages state and performs computations for an operation and its strategy.\n\nWorkers are spawned by strategies to manage state and perform computations for a given\noperation. Any operation or strategy state not stored in the deployment lives in a worker.\n\nSkitter workers are created by strategies with an initial state. Any messages received by the\nworker are handled by the `c:Skitter.Strategy.Operation.process/4` hook of its strategy. This\ncallback receives the current worker state and may return a new, updated state to be stored by\nthe worker.\n\nSince strategies can create many separate workers, each worker is created with a _role_ which\ncan be used by the strategy to provide different implementations of\n`c:Skitter.Strategy.Operation.process/4` based on the worker that received the message.\n\nThis module defines the worker types and various functions to deal with workers.","ref":"Skitter.Worker.html"},{"type":"function","title":"Skitter.Worker.create_local/3","doc":"Create a new worker on the local node.\n\nThis will raise when executed on a master node.","ref":"Skitter.Worker.html#create_local/3"},{"type":"function","title":"Skitter.Worker.create_remote/4","doc":"Create a new worker on a remote node.\n\nThe worker will be placed on a random node, subject to the passed placement constraints.","ref":"Skitter.Worker.html#create_remote/4"},{"type":"function","title":"Skitter.Worker.send/2","doc":"Send a message to the worker at `ref`.","ref":"Skitter.Worker.html#send/2"},{"type":"function","title":"Skitter.Worker.stop/1","doc":"Stop worker `ref`.","ref":"Skitter.Worker.html#stop/1"},{"type":"type","title":"Skitter.Worker.placement/0","doc":"Placement constraints.\n\nWhen spawning a remote worker, it is often desirable to tweak on which node the worker will be\nplaced.  This type defines a set of placement constraints which can be passed as an argument to\n`create_remote/4`.\n\nThe following constraints are defined:\n\n- `nil`: No constraints.\n- `on: node`: Spawn the worker at the specified node.\n- `with: ref`: Spawn the worker on the same node as the worker identified by `ref`.\n- `avoid: ref`: Try to place the worker on a different node than the worker identified by `ref`.\n- `avoid: node`: Try to avoid placing the worker on `node`.\n- `tagged: tag`: Try to place the worker on a node with a specific `t:Skitter.Nodes.tag/0`.\n\nNote that it is not always possible to match the desired constraints. When this is the case, a\nwarning will be logged.","ref":"Skitter.Worker.html#t:placement/0"},{"type":"type","title":"Skitter.Worker.ref/0","doc":"Reference to a created worker.","ref":"Skitter.Worker.html#t:ref/0"},{"type":"type","title":"Skitter.Worker.role/0","doc":"Worker role.\n\nEach worker is tagged with a role which allows the strategy to differentiate between the various\nworkers it creates.","ref":"Skitter.Worker.html#t:role/0"},{"type":"type","title":"Skitter.Worker.state/0","doc":"Worker state.","ref":"Skitter.Worker.html#t:state/0"},{"type":"type","title":"Skitter.Worker.state_or_state_fn/0","doc":"Worker state or a function which returns a worker state.\n\nFunctions which create workers may be created with an initial state, or with a function which\nreturns an initial state. The function is executed inside the worker when it is first created.","ref":"Skitter.Worker.html#t:state_or_state_fn/0"},{"type":"module","title":"Skitter.BIO","doc":"Shorthands for using built-in operations.\n\nThis module defines various macros which can be used to use the various built-in operations in\nskitter workflows.","ref":"Skitter.BIO.html"},{"type":"macro","title":"Skitter.BIO.filter/2","doc":"`Skitter.BIO.Filter` node.\n\nInserts a `Skitter.BIO.Filter` `Skitter.DSL.Workflow.node/2` in the workflow. The argument\npassed to this macro is passed as an argument to `Skitter.BIO.Filter`, other options (`as:`,\n`with:`) should be passed as a second, optional argument.","ref":"Skitter.BIO.html#filter/2"},{"type":"macro","title":"Skitter.BIO.flat_map/2","doc":"`Skitter.BIO.FlatMap` node.\n\nLike `map/2`, but with `Skitter.BIO.FlatMap`.","ref":"Skitter.BIO.html#flat_map/2"},{"type":"macro","title":"Skitter.BIO.keyed_reduce/4","doc":"`Skitter.BIO.KeyedReduce` node.\n\nInserts a `Skitter.BIO.KeyedReduce` `Skitter.DSL.Workflow.node/2` in the workflow. The `key_fn`,\n`red_fn` and `initial` arguments passed to this macro are passed as arguments to\n`Skitter.BIO.KeyedReduce`. Other options (`as:`, `with:`) can be passed as a fourth argument.","ref":"Skitter.BIO.html#keyed_reduce/4"},{"type":"macro","title":"Skitter.BIO.map/2","doc":"`Skitter.BIO.Map` node.\n\nInserts a `Skitter.BIO.Map` `Skitter.DSL.Workflow.node/2` in the workflow. The argument passed\nto this macro is passed as an argument to `Skitter.BIO.Map`, other options (`as:`, `with:`)\nshould be passed as a second, optional argument.","ref":"Skitter.BIO.html#map/2"},{"type":"macro","title":"Skitter.BIO.msg_source/1","doc":"Message source node.\n\nInserts a `Skitter.BIO.MessageSource` node in the workflow. Any options are passed to the\nworkflow.","ref":"Skitter.BIO.html#msg_source/1"},{"type":"macro","title":"Skitter.BIO.print/2","doc":"`Skitter.BIO.Print` node.\n\nInsert a `Skitter.BIO.Print` node in the workflow. The argument passed to this macro is passed\nas the print label described in the operation documentation. Workflow options (`as`, `with`) can\nbe passed as the optional second argument.","ref":"Skitter.BIO.html#print/2"},{"type":"macro","title":"Skitter.BIO.send_sink/2","doc":"`Skitter.BIO.Send` node.\n\nInsert a `Skitter.BIO.Send` sink in the workflow. The argument passed to this macro is passed\nas the pid described in the operation documentation. Workflow options (`as`, `with`) can\nbe passed as the optional second argument. When no argument is provided, `self()` will be used.","ref":"Skitter.BIO.html#send_sink/2"},{"type":"macro","title":"Skitter.BIO.stream_source/2","doc":"Stream source node.\n\nInserts a `Skitter.BIO.StreamSource` node in the workflow. The provided `enum` is passed as an\nargument to `Skitter.BIO.StreamSource`. `opts` are passed as options to the workflow.","ref":"Skitter.BIO.html#stream_source/2"},{"type":"macro","title":"Skitter.BIO.tcp_source/3","doc":"Tcp source node.\n\nInserts a `Skitter.BIO.TCPSource` node in the workflow. The address and ports passed to this\nargument will be passed as arguments to `Skitter.BIO.TCPSource`. Provided options are passed to\nthe workflow.","ref":"Skitter.BIO.html#tcp_source/3"},{"type":"module","title":"Skitter.BIO.Filter","doc":"Filter operation.\n\nThis operation implements a filter. When embedded inside a workflow, this operation is provided\nwith a function argument. This function will be called to process every element received by the\noperation. If the function returns true, the received element will be sent to the `accept` port.\nOtherwise, the element will be sent to the `reject` out port.","ref":"Skitter.BIO.Filter.html"},{"type":"module","title":"Operation Properties - Skitter.BIO.Filter","doc":"* in ports: `_`\n* out ports: `accept, reject`\n* default strategy: `Skitter.BIS.ImmutableLocal`","ref":"Skitter.BIO.Filter.html#module-operation-properties"},{"type":"module","title":"Skitter.BIO.FlatMap","doc":"FlatMap operation.\n\nThis operation implements a flatmap. When embedded inside a workflow, this operation is provided\nwith a function argument. This function will be called to process every element received by the\noperation. This function should return a list. Each element in this list will be emitted on the\n`_` out port.","ref":"Skitter.BIO.FlatMap.html"},{"type":"module","title":"Operation Properties - Skitter.BIO.FlatMap","doc":"* in ports: `_`\n* out ports: `_`\n* default strategy: `Skitter.BIS.ImmutableLocal`","ref":"Skitter.BIO.FlatMap.html#module-operation-properties"},{"type":"module","title":"Skitter.BIO.KeyedReduce","doc":"Keyed Reduce operation.\n\nThis operation implements a reduce operation. It accepts three arguments wrapped in a tuple when\nembedded inside a workflow: a key function, a reduce function and an initial state. When this\noperation receives data, the key function is called with the received data as its first\nargument. The key function should return a key, which is used to obtain the state associated\nwith the key. Afterwards, the reduce function is called with the received data as its first\nargument and the state associated with the key returned by the key function as its second\nargument. This function should return a `{state, emit}` tuple. The first value of this tuple\nwill be stored as the new state of the key, while the second value of this tuple will be emitted\non the `_` out port.","ref":"Skitter.BIO.KeyedReduce.html"},{"type":"module","title":"Operation Properties - Skitter.BIO.KeyedReduce","doc":"* in ports: `_`\n* out ports: `_`\n* default strategy: `Skitter.BIS.KeyedState`","ref":"Skitter.BIO.KeyedReduce.html#module-operation-properties"},{"type":"module","title":"Skitter.BIO.Map","doc":"Map operation.\n\nThis operation implements a map. When embedded inside a workflow, this operation is provided\nwith a function argument. This function will be called to process every element received by the\noperation. The result of applying the function to the received data will be emitted on the `_`\nout port.","ref":"Skitter.BIO.Map.html"},{"type":"module","title":"Operation Properties - Skitter.BIO.Map","doc":"* in ports: `_`\n* out ports: `_`\n* default strategy: `Skitter.BIS.ImmutableLocal`","ref":"Skitter.BIO.Map.html#module-operation-properties"},{"type":"module","title":"Skitter.BIO.MessageSource","doc":"Development source which can receive messages from elixir processes.\n\nThis operation is a source which can receive data from Elixir messages. After the operation is\ndeployed, `send/2` can be used to send a message to the source. Only one message source can\nexist at the same time.\n\nThis operation is only intended to be used locally. It does not work in a distributed\nenvironment.","ref":"Skitter.BIO.MessageSource.html"},{"type":"module","title":"Operation Properties - Skitter.BIO.MessageSource","doc":"* in ports: none\n* out ports: `_`\n* default strategy: `Skitter.BIS.PassiveSource`","ref":"Skitter.BIO.MessageSource.html#module-operation-properties"},{"type":"function","title":"Skitter.BIO.MessageSource.send/1","doc":"Send a message to the message source.","ref":"Skitter.BIO.MessageSource.html#send/1"},{"type":"module","title":"Skitter.BIO.Print","doc":"Print operation.\n\nThis operation print any data element it receives and emits the received value on its out port.\nThis makes it possible to insert this operation in the middle of a data processing\npipeline for debugging purposes.\n\nA single string should be provided as an argument, this string will be used as a prefix for the\nprinted output.","ref":"Skitter.BIO.Print.html"},{"type":"module","title":"Operation Properties - Skitter.BIO.Print","doc":"* in ports: `_`\n* out ports: `_`\n* default strategy: `Skitter.BIS.ImmutableLocal`","ref":"Skitter.BIO.Print.html#module-operation-properties"},{"type":"module","title":"Skitter.BIO.Send","doc":"Send operation.\n\nThis operation is a sink that sends any data it receives to a given pid. The pid that should\nreceive the messages should be provided as an argument when the operation is embedded inside a\nworkflow.","ref":"Skitter.BIO.Send.html"},{"type":"module","title":"Operation Properties - Skitter.BIO.Send","doc":"* in ports: `_`\n* out ports: none\n* default strategy: `Skitter.BIS.ImmutableLocal`","ref":"Skitter.BIO.Send.html#module-operation-properties"},{"type":"module","title":"Skitter.BIO.StreamSource","doc":"Source which produces a stream of predefined data.\n\nThis operation is a source which is created with a set of data (an Elixir `Enumerable`). Once\ndeployed, it will emit each element in the enumerable in order.","ref":"Skitter.BIO.StreamSource.html"},{"type":"module","title":"Operation Properties - Skitter.BIO.StreamSource","doc":"* in ports: none\n* out ports: `_`\n* default strategy: `Skitter.BIS.StreamSource`","ref":"Skitter.BIO.StreamSource.html#module-operation-properties"},{"type":"module","title":"Skitter.BIO.TCPSource","doc":"TCP source operation.\n\nThis operation is a source which listens to data on a tcp socket. This operation should be\nembedded in the workflow with a keyword list as its argument. This list should contain `address`\nand `port` keys, which specify the address and port to connect to, respectively.\n\nA data element is emitted for each line sent to the tcp socket.","ref":"Skitter.BIO.TCPSource.html"},{"type":"module","title":"Operation Properties - Skitter.BIO.TCPSource","doc":"* in ports: none\n* out ports: `_`\n* default strategy: `Skitter.BIS.PassiveSource`","ref":"Skitter.BIO.TCPSource.html#module-operation-properties"},{"type":"module","title":"Skitter.BIS.ImmutableLocal","doc":"Strategy for stateless operations.\n\nThis strategy can be used for stateless operations. It expects an operation to provide a `react`\ncallback. When the operation receives data, a worker on the current worker node is selected.\nThe selected worker will call the react callback with the received data.","ref":"Skitter.BIS.ImmutableLocal.html"},{"type":"module","title":"Operation Properties - Skitter.BIS.ImmutableLocal","doc":"* in ports: A single in port is required.\n* out ports: This strategy places no limitations on the out ports of the operation.\n* callbacks:\n  * `react` (required): Called for each incoming data element.\n  * `conf` (optional): Called at deployment time. The result will be passed as config to each\n  invocation of `react`.","ref":"Skitter.BIS.ImmutableLocal.html#module-operation-properties"},{"type":"module","title":"Skitter.BIS.KeyedState","doc":"Strategy for stateful operations that can partition their state by key.\n\nThis strategy can be used for stateful operations with a keyed state. It expects an operation to\nprovided a `key` and a `react` callback. When the operation receives data, the `key` callback\nwill be called to determine the key of the incoming data, afterwards, the value will be sent to\na worker maintaining the state for key. This worker might live on another cluster node. Finally,\nthis worker will call the `react` callback to update the state of the key.","ref":"Skitter.BIS.KeyedState.html"},{"type":"module","title":"Operation Properties - Skitter.BIS.KeyedState","doc":"* in ports: A single in port is required.\n* out ports: This strategy places no limitations on the out ports of the operation.\n* callbacks:\n  * `key` (required): Called for each incoming data element. Can not access operation state.\n  * `react` (required): Called for each incoming data element.\n  * `conf` (optional): Called to create a configuration for the operation.\n  * `initial_state` (optional): Called to create the initial state for a key when it occurs for\n    the first time. If it is not provided, the state defaults to `nil`.","ref":"Skitter.BIS.KeyedState.html#module-operation-properties"},{"type":"module","title":"Skitter.BIS.PassiveSource","doc":"Strategy for reactive source operations.\n\nThis strategy can be used to create a source operation. It is designed for operations that\nsend data into the workflow as a response to incoming data.\n\nWhen the operation is deployed, the strategy will spawn a single worker and call the operation's\n`subscribe` callback with the arguments provided in the workflow. This callback should ensure\nthat messages are sent to the worker in response to external events. Any time a message is sent\nto the worker, the strategy will call the `process` callback of the operation, the operation\nshould emit a (list of) data elements based on the received message.\n\nThe strategy ensures that `process` is called on a random worker node, to evenly distribute the\nincoming data over the cluster.","ref":"Skitter.BIS.PassiveSource.html"},{"type":"module","title":"Operation Properties - Skitter.BIS.PassiveSource","doc":"* in ports: none\n* out ports: This strategy places no limitations on the out ports of the operation.\n* callbacks:\n  * `subscribe` (required): Called at deployment time with the workflow arguments. This callback\n  should ensure the worker receives messages in response to event.\n  * `process` (required): Called for each received message. This callback should emit the list\n  of received data to its out port to push them into the workflow.","ref":"Skitter.BIS.PassiveSource.html#module-operation-properties"},{"type":"module","title":"Skitter.BIS.StreamSource","doc":"Strategy for stream-based source operations.\n\nThis strategy can be used to create a source operation. It is designed for operations which\ngenerate a stream of data that is to be sent into the workflow.\n\nWhen the operation is deployed, this strategy will spawn a single worker and call the operations\n`stream` callback. This callback should return a stream. Once deployed, the elements of this\nstream will be emitted by one by one. The strategy ensures these values are shuffled over the\navailable worker nodes.","ref":"Skitter.BIS.StreamSource.html"},{"type":"module","title":"Operation Properties - Skitter.BIS.StreamSource","doc":"* in ports: none\n* out ports: a single out port.\n* callbacks:\n  * `stream`: Called at deployment time. This callback should return a stream, which will be\n  emitted once the operation has been deployed.","ref":"Skitter.BIS.StreamSource.html#module-operation-properties"},{"type":"module","title":"Skitter.Dot","doc":"Export skitter workflows as [graphviz](https://graphviz.org/) dot graphs.\n\nThe main function in this module is the `to_dot/1` function, which accepts a skitter workflow\nand returns its dot representation as a string. End users may prefer the `print_dot/1` function,\nwhich immediately prints the returned string. If dot is installed on the system, the `export/3`\nfunction can be used to export the generated graph in a variety of formats.","ref":"Skitter.Dot.html"},{"type":"function","title":"Skitter.Dot.render/3","doc":"Renders the generated dot graph, requires dot to be installed on the system.\n\nThis function exports a given workflow to the dot language (using `to_dot/1`), after which it\ncalls `dot` on the generated dot representation. When `dot` returns successfully, `{:ok,\nstring}` is returned, where `string` is the output generated by `dot`.\n\nA `format` should be specified. This format is passed to `dot` through the use of its `-T`\noption. For a list of the options supported on your system, see `man dot`.\n\nAn optional list of options may be passed to further configure the use of `dot`. The following\noptions are supported:\n- `dot_exe`: the path to the dot executable, by default, this function assumes dot is present in\nyour `$PATH`.\n- `extra`: a list of extra arguments to pass to the dot executable.","ref":"Skitter.Dot.html#render/3"},{"type":"function","title":"Examples - Skitter.Dot.render/3","doc":"Save `workflow` as a svg file:\n```\nrender(workflow, \"svg\")\n```","ref":"Skitter.Dot.html#render/3-examples"},{"type":"function","title":"Skitter.Dot.render_to_file/4","doc":"Renders the generated dot graph and save it to `path`.\n\nRenders the provided workflow (with `render/3`) and store the output to `path`. `format` and\n`opts` are passed to `render/3`.","ref":"Skitter.Dot.html#render_to_file/4"},{"type":"function","title":"Skitter.Dot.to_dot/1","doc":"Return the dot representation of a workflow as a string.","ref":"Skitter.Dot.html#to_dot/1"},{"type":"module","title":"Skitter.ExitCodes","doc":"Functions which return the various exit codes a Skitter application may return.","ref":"Skitter.ExitCodes.html"},{"type":"function","title":"Skitter.ExitCodes.remote_shutdown/0","doc":"Returned when the runtime shut down because connection to a remote runtime was lost","ref":"Skitter.ExitCodes.html#remote_shutdown/0"},{"type":"module","title":"Skitter.Release","doc":"Utilities for building Skitter releases.\n\nReleases (`mix release`) are used to deploy Skitter applications over a cluster. Skitter\nincludes several configuration files and scripts to customize the generated release and to\nfaciliate its distribution. In order for your application to use these scripts, your release\nconfiguration should use the `step/1` function defined in this module:\n\n```\ndef project do\n  ...,\n  releases: [\n     : [steps: [&Skitter.Release.step/1, :assemble]]\n  ],\n  ...\nend\n```\n\nThis will cause Skitter to customize the configuration of your release to include the required\nconfiguration files and scripts.\n\n> #### `mix skitter.new` {:.info}\n>\n> Releases are already configured if `mix skitter.new` was used to create the Skitter project.\n\nPlease refer to the [deployment page](deployment.html#releases) for more information about\nSkitter releases and how to use them.","ref":"Skitter.Release.html"},{"type":"function","title":"Skitter.Release.step/1","doc":"Customize the release assembly.\n\nThis function customizes the configuration of a release to include several Skitter configuration\nfiles and scripts. Concretely, it makes the following modifications:\n\n- Modify the release to only include unix executables.\n- Add Skitter-specific `env.sh.eex` and `vm.args` files.\n- Add the `skitter` deploy script used to manage Skitter runtime systems.\n- Adds a `skitter.exs` script which is used to configure the Skitter runtime before it starts.\n\nThis function should be added to the `steps` configuration of your release __before__ the\n`:assemble` step. It is best to include it after any other steps, as this enables the function\nto verify that the Skitter generated configuration does not overwrite user-provided\nconfiguration values.\n\nThe following assumptions about your release configuration are made:\n\n- The `:include_executables_for` option is not set. Instead, it is set to `[:unix]`.\n- The `:rel_templates_path` option is not set. Instead, it is set to point to a directory\n  containing Skitter specific `env.sh.eex` and `vm.args.eex` files.","ref":"Skitter.Release.html#step/1"},{"type":"exception","title":"Skitter.DefinitionError","doc":"Raised when an error is encountered in a Operation, Workflow or Strategy.\n\nThis error can be raised by a Skitter DSL or by a skitter data type. It is raised when invalid\nsyntax is encountered or when a semantic error is present in a `Skitter.Operation`,\n`Skitter.Workflow` or `Skitter.Strategy`.","ref":"Skitter.DefinitionError.html"},{"type":"exception","title":"Skitter.StrategyError","doc":"Error raised by a strategy.\n\nThis error can be raised by a `Skitter.Strategy`. It is used to indicate the strategy\nencountered some unexpected situation.","ref":"Skitter.StrategyError.html"},{"type":"extras","title":"Overview","doc":"# Overview\n\nSkitter is a distributed stream processing framework written in Elixir which\nenables the creation of scalable, distributed stream processing applications.\nUnique to Skitter is its ability to control how the various operations which\nconstitute a stream processing application are distributed at runtime.\n\nThese pages serve as the main documentation for Skitter users. They consist of\nthree main parts, which can be accessed from the top of the sidebar.\n\n* The __manual__: this is the section of the docs you are currently browsing.\n  It is a collection of guides which document how to get started with Skitter.\n  We recommend first-time users to read through most pages in this manual to\n  get familiar with Skitter.\n\n* The __modules__ section: the pages in this section contain the detailed\n  documentation of the API offered by Skitter. Various pages in the manual will\n  link to the pages in this section of the documentation. We recommend to\n  browse these pages on an as-needed basis.\n\n* The __mix tasks__ section: the pages in this section document the mix tasks\n  defined by Skitter. `Mix` is the build tool used by Elixir. Mix tasks are\n  small command line programs which are used to extend the functionality of\n  mix. We recommend browsing these pages only when needed.\n\nThe remainder of this manual is, itself, divided into several parts:\n\n* The __getting started__ pages include this page and the installation\n  instructions. They are intended to help you install Skitter, create your\n  first Skitter application and execute it.\n\n* The __concepts__ pages explain the various language constructs introduced by\n  Skitter and how they are used to define a distributed stream processing\n  application. These pages form the main part of this manual.\n\n* The __deployment__ pages document how Skitter applications are deployed over\n  a cluster and how they are configured. These pages can be skipped until you\n  are ready to run your application in a distributed setting.\n\n* The __guides__ pages document various niche aspects of Skitter. These\n  should only need be read when needed.\n\nThroughout this manual, some knowledge of Elixir is assumed. The\n[official guide](https://elixir-lang.org/getting-started/introduction.html) and\nthe [official recommended resources](https://elixir-lang.org/learning.html) are\ngreat places to familiarise yourself with Elixir if needed.","ref":"overview.html"},{"type":"extras","title":"Installation","doc":"# Installation","ref":"installation.html"},{"type":"extras","title":"Elixir - Installation","doc":"Skitter is developed as a DSL in [Elixir](https://elixir-lang.org). Therefore,\nElixir is required to write and run Skitter applications. In turn, Elixir is\nbuilt upon Erlang/OTP. Therefore, an Erlang/OTP installation is also required.\nSkitter requires Elixir version 1.16 or later running on Erlang/OTP version 26\nor later.\n\nThe [official Elixir documentation](https://elixir-lang.org/install.html)\nprovides instructions on how to install Elixir (and Erlang/OTP).","ref":"installation.html#elixir"},{"type":"extras","title":"Skitter - Installation","doc":"We provide an application generator which is used to create a new Skitter\nproject. This generator creates a new Elixir project, configures it to use\nSkitter and provides some initial code to help you get started.\n\nThe application generator is provided as a mix task. Mix is an extendable build\ntool used by Elixir; it should have been installed together with Elixir.\nIn order to use the application generator, you need to add it to your local mix\ninstallation, which can be done as follows:\n\n```\n$ mix archive.install hex skitter_new\n```\n\nThis fetches the `skitter_new` project from hex, Elixir's package repository,\nand adds it to your local `mix` installation.\n\nOnce the task is installed, you can run `mix help skitter.new` to see if\neverything is installed correctly. It should produce text detailing how to use\nthe task.","ref":"installation.html#skitter"},{"type":"extras","title":"Up and Running","doc":"# Up and Running","ref":"up_and_running.html"},{"type":"extras","title":"Creating a project - Up and Running","doc":"Once [installed](installation.html#skitter), `mix skitter.new  `\ncan be used to create a new Skitter project. This task creates a new Elixir\nproject, configures it to use Skitter and provides some initial code to help\nyou get started.\n\n`mix skitter.new` can be used from any directory to create a new Skitter\nproject.  The generator expects a `snake_case` project name; i.e. words should\nbe lower cased and spaces should be replaced by underscores. For the purposes\nof this guide, we will create a project called `hello_skitter` and use it as a\nrunning example.\n\nThe project can be created with the following command:\n```\n$ mix skitter.new hello_skitter\n```\n\n> #### mix task documentation {:.tip}\n>\n> Mix tasks accept command line flags and options that may tweak their\n> behaviour. To find out more, you can click on any `mix  ` block in\n> these docs to read the task's documentation.\n>\n> For instance, you can click on `mix skitter.new` to learn more about its\n> behaviour.\n\n> #### Skitter without `mix skitter.new` {:.info}\n>\n> It is possible to add Skitter as a dependency to an existing Elixir project.\n> However, we recommend the use of `mix skitter.new`, especially for new users.\n>\n> Users who wish to add Skitter to an exising project should add `:skitter` as\n> a dependency and customize their [release configuration](`mix release`) as\n> detailed in `Skitter.Release`.\n\nThe tool will create the project:\n\n```text\n* creating hello_skitter\n* creating hello_skitter/lib\n* creating hello_skitter/config\n* creating hello_skitter/mix.exs\n* creating hello_skitter/.iex.exs\n* creating hello_skitter/.formatter.exs\n* creating hello_skitter/config/config.exs\n* creating hello_skitter/lib/hello_skitter.ex\n* creating hello_skitter/.gitignore\n* creating hello_skitter/README.md\n```\n\nAfter which it will ask if it should fetch and compile the dependencies of the\nproject. Either reply `y`, or `cd` into the project directory and run `mix\ndeps.get` followed by `mix deps.compile`.\n\n```text\nFetch and build dependencies? [Yn] y\n* Running `mix deps.get` in hello_skitter\n \n* Running `mix deps.compile` in hello_skitter\n \n\nGenerated skitter app\n\n  Your skitter project has been created at `hello_skitter`.\n  You can now start working on your Skitter application:\n\n  $ cd hello_skitter\n  $ iex -S mix\n\n  For your convenience, the generated README.md file contains a\n  summary of the generated files and a summary of elixir commands.\n```\n\nAs the generator says, we can `cd` into the directory of the project:\n\n```shell\n$ cd hello_skitter\n```\n\nand execute the project in the elixir shell:\n\n```shell\n$ iex -S mix\n```\n\nThis will start `iex`, the elixir shell, and start the current mix project\n(i.e.  `hello_skitter`). The generator generated an example \"word count\"\napplication and configured Skitter to automatically deploy the application when\nstarted.  Therefore, starting the project will produce the following output:\n\n```text\nErlang/OTP 26 [erts-14.2.2] [source] [64-bit] [smp:16:16] [ds:16:16:10] [async-threads:1] [jit:ns] [dtrace]\n\n[19:29:31.329][info] ⬡⬢⬡⬢ Skitter v0.7.0 started in local mode\n[19:29:31.343][info] Deploying &HelloSkitter.workflow/0\n{\"Hello\", 1}\n{\"Skitter\", 1}\n{\"Hello\", 2}\n{\"World!\", 1}\nInteractive Elixir (1.16.2) - press Ctrl+C to exit (type h() ENTER for help)\niex(1)>\n```\n\nTo exit `iex`, press `Ctrl + c` twice.\n\nYou now have a working Skitter project. The project generator created a\n`README.md` file in your project, which tells you about the various files it\ngenerated. Comments are also present in the various generated files. The\nfollowing section of the Skitter manual introduces the concepts required to\nwrite Skitter applications of your own. Before we end this guide, we briefly\ndiscuss the various ways in which you can run a Skitter project and the\nstructure of an Elixir project.","ref":"up_and_running.html#creating-a-project"},{"type":"extras","title":"Running the project - Up and Running","doc":"The most common way to run an elixir application is to use `iex -S mix`. `iex`\nloads the Elixir shell, `iex`, while `-S mix` tells it to start `mix`, the\nElixir build tool, which will load your Skitter project.\n\n> #### iex {:.info}\n>\n> The [official documentation](`IEx`) contains of useful information on how to\n> use `iex` and on how it can be customized.\n\nIf you wish to run your application without starting a shell, you can use\n`mix run --no-halt`, which will start your mix project without starting `iex`.\nAs before, `Ctrl + c` can be used to exit your application.\n\n```text\n$ mix run --no-halt\n[19:53:22.826][info] ⬡⬢⬡⬢ Skitter v0.7.0 started in local mode\n[19:53:22.841][info] Deploying &HelloSkitter.workflow/0\n{\"World!\", 1}\n{\"Skitter\", 1}\n{\"Hello\", 1}\n{\"Hello\", 2}\n```\n\nOnce you have a working application, you can run it on a cluster computer.\nSkitter supports executing your application on a cluster (i.e. on several\ncomputers) and also enables the simulation of a distributed system on a single\nmachine. Please refer to the [deployment documentation](deployment.html) for\nmore information about both of these topics.","ref":"up_and_running.html#running-the-project"},{"type":"extras","title":"Elixir project structure - Up and Running","doc":"`mix skitter.new` generates a `README.md` file which documents the overall\nstructure of the application it generated. We recommend reading this README and\nbrowsing the generated files to see how everything fits together. We provide a\nshort summary of the most important files and directories in the project below:\n* `mix.exs`: Configures how `mix`, the elixir build tool, runs and compiles\n  your application.\n* `config/config.exs`: Configures your application and its dependencies (e.g.\n  skitter, logger).\n* `lib/`: contains the code of your application. Any `.ex` file present in\n  this directory will be picked up by mix, compiled and included in your\n  project.","ref":"up_and_running.html#elixir-project-structure"},{"type":"extras","title":"Overview","doc":"# Overview\n\nSkitter is a distributed stream processing framework. The idea behind these\nframeworks is to process various _streams_ of incoming data. Distributed stream\nprocessing frameworks are designed to process large volumes of data. This is\ndone by _distributing_ the processing of these streams over different computers.\n\nSimilar to other distributed stream processing frameworks, Skitter offers a\nprogramming model where streams of data are processed through various\n**operations**. An operation is connected to several streams and is responsible\nfor processing data that arrives on these streams. The operation may also emit\ndata on streams, which can then be processed by other operations in the\napplication.\n\n![](assets/docs/operations.svg)\n\nOperations which do not receive input and only produce values are called\n*sources*, while operations which receive input and do not produce any output\nare called *sinks*.\n\nA Skitter application is created by combining several of\nthese operations into a **workflow**.\n\n![](assets/docs/workflow.svg)\n\nOnce created, a Skitter workflow can be distributed over a cluster (i.e. over a\ncollection of different computers), after which it can start to process\nincoming data records.\n\nExisting distributed stream processing frameworks automatically distribute a\nstream processing application over a cluster. Unique to Skitter is its support\nfor specifying how each operation in a stream processing application is\ndistributed. This is done by writing a **distribution strategy**.","ref":"concepts.html"},{"type":"extras","title":"Three languages - Overview","doc":"Therefore, Skitter introduces three languages:\n\n- The **workflow** language is used to combine several existing operations into\n  a workflow.\n- The **operation** language is used to create new operations which can be used\n  in these workflows.\n- The **strategy** language is used to define how operations can be distributed\n  over a cluster.\n\nEach of these languages is introduced in detail in the following pages of this\nmanual.","ref":"concepts.html#three-languages"},{"type":"extras","title":"Workflows","doc":"# Workflows\n\nSkitter applications are defined as the combination of several data processing\nsteps connected to one another. We call such a combination of connected data\nprocessing steps a _workflow_.\n\nEach data processing step is defined as an _operation_, which we discuss in the\nnext section of this guide. For now, it suffices to think of these operations\nas black boxes which ingest data, potentially emitting new data as they do so.\nAn operation receives data on its input _ports_ and emits data using its output\nports. A workflow combines several of these operations by connecting the out\nports of operations to the in ports of other operations.\n\nAs an example, let's define the \"Hello, World!\" application of stream\nprocessing frameworks: a word count:\n\n![a schematic depiction of a word count application consisting of two nodes: split and count](assets/docs/word_count_schematic.svg)\n\nThis workflow consists of four operations: source, split and print:\n- The source operation produces the sentences of which the words will be\n  counted.\n- The split operation accepts sentences and splits them, emitting individual\n  words.\n- The count operation accepts words and returns `(word, count)` pairs, which\n  specify how many times `word` has been seen thus far.\n- The print operation accepts the count pairs and shows them to the user.\n\nLet's assume that we have defined a Skitter operation for each of these steps,\ncalled `Source`, `Split`, `Count` and `Print`. We can now define the word count\napplication as a Skitter workflow:\n\n```\nuse Skitter\n\nword_count = workflow do\n  node(Source, as: source)\n  node(Split, as: split)\n  node(Count, as: count)\n  node(Print, as: print)\n\n  source.out ~> split.sentences\n  split.words ~> count.words\n  count.pairs ~> print.pairs\nend\n```\n\nThe above code uses `Skitter.DSL.Workflow.workflow/2` to define a workflow. A\nworkflow is defined by combining several _nodes_, each of which \"wrap\" an\noperation, and by \"wiring\" these nodes together through the use of _links_.\nNodes are created with the use of the `Skitter.DSL.Workflow.node/2` macro,\nwhile links are introduced through the use of `Skitter.DSL.Workflow.~>/2`.\n\nVisually, this workflow can be represented as follows:\n\n![Graphical representation of the workflow shown above](assets/docs/word_count.svg)\n\n> #### Visualising workflows {:.tip}\n>\n> Skitter workflows can be visualised through the use of the functions defined\n> in `Skitter.Dot`. The visual representations of workflows shown in this\n> document were generated through the use of this tool.","ref":"workflows.html"},{"type":"extras","title":"Nodes - Workflows","doc":"Nodes in a workflow represent a single data processing step. They typically\nrefer to an operation, which implements the data processing logic of the node,\nand to a strategy, which determines how the operation is distributed over the\ncluster. A node also has a name, which is used to refer to the node when\ncreating links. Nodes may also accept arguments, which can be used to provide\ninitialization arguments to the strategy or operation of the node.\n\nWhen a node is created, an operation must be provided, the name, strategy and\narguments of the node may be provided, through the use of `as:`, `with:`, and\n`args:`, respectively.\n\n```\nnode(SomeOperation, with: someStrategy, args: [some_argument], as: name)\n```\n\nWhen no name is provided, the workflow language will generate a unique name.\nIf no strategy is present, the _default strategy_ of the operation will be\nused. Arguments default to `nil` when not provided.","ref":"workflows.html#nodes"},{"type":"extras","title":"Links and syntactic sugar - Workflows","doc":"Links in a workflow connect a particular output of a node with the input of\nanother node.\n\n```\noperation.out_port ~> other_operation.in_port\n```\n\nThe link operator, `Skitter.DSL.Workflow.~>/2`, introduces syntactic sugar to\nmake workflow definitions more concise. When the left-hand side of `~>`\noperator contains a node definition, the first output port of this node will be\nconnected to the right-hand side of the link. Similarly, when the right-hand\nside of `~>` contains a node definition, the left-hand side of the link will be\nconnected to the first input of this node. This allows node definitions in a\nworkflow to be chained:\n\n```\nuse Skitter\n\nword_count = workflow do\n  node(Source)\n  ~> node(Split)\n  ~> node(Count)\n  ~> node(Print)\nend\n```\n\n![Graphical representation of the workflow shown above (it remains largely unchanged)](assets/docs/word_count_chained.svg)","ref":"workflows.html#links-and-syntactic-sugar"},{"type":"extras","title":"Operator-style workflow definitions - Workflows","doc":"Many distributed stream-processing frameworks offer a programming model where\napplications are created by chaining various _operators_, such as `map` and\n`reduce`. This can also be done in Skitter:\n\n```\nuse Skitter\n\nword_count = workflow do\n  stream_source(~w(Hello Skitter Hello World!))\n  ~> flat_map(&String.split/1)\n  ~> keyed_reduce(fn word -> word end, fn word, ctr -> {ctr + 1, {word, ctr + 1}} end, 0)\n  ~> print()\nend\n```\n\n![Graphical representation of the operator-style workflow definition](assets/docs/word_count_operators.svg)\n\nThis is possible because Skitter predefines various operations, called\n_built-in operations_, BIO for short. The full list of BIOs can be found in the\nmodules section of the documentation. In the example above, the following BIOs\nare used:\n\n- `Skitter.BIO.StreamSource`\n- `Skitter.BIO.FlatMap`\n- `Skitter.BIO.KeyedReduce`\n- `Skitter.BIO.Print`\n\nThese BIOs can be used in a workflow like regular nodes:\n\n```\nuse Skitter\n\nword_count = workflow do\n  node(Skitter.BIO.StreamSource, args: ~w(Hello Skitter Hello World!))\n  ~> node(Skitter.BIO.FlatMap, args: &String.split/1)\n  ~> node(Skitter.BIO.KeyedReduce, args: {fn word -> word end, fn word, ctr -> {ctr + 1, {word, ctr + 1}} end, 0})\n  ~> node(Skitter.BIO.Print)\nend\n```\n\nHowever, the `Skitter.BIO` module defines various macros which can be used to\ndefine these nodes, as shown earlier in this section. The full list of these\nmacros is defined in `Skitter.BIO`. Since Skitter is extensible, you are able\nto add your own operators to Skitter. How this is done is defined in the\n[operators guide](operators.md).","ref":"workflows.html#operator-style-workflow-definitions"},{"type":"extras","title":"Nested workflows - Workflows","doc":"Finally, workflows can be stored as values and used inside nodes. This makes it\npossible to reuse workflows:\n\n```\nuse Skitter\n\ndefmodule Example do\n  def listen_and_preprocess(port) do\n    workflow out: integers do\n      tcp_source(\"127.0.0.1\", port) ~> map(&String.to_integer/1) ~> integers\n    end\n  end\n\n  def workflow do\n    workflow do\n      node(listen_and_preprocess(5555), as: top) ~> join.left\n      node(listen_and_preprocess(5556), as: bottom) ~> join.right\n\n      node(Join, as: join)\n      ~> print()\n    end\n  end\nend\n```\n\n`Example.workflow()` will return the following workflow:\n\n![Graphical representation of the nested workflows](assets/docs/workflow_nested.svg)\n\nNested workflows are always inlined before a workflow is deployed over the\ncluster.\n\n![Graphical representation of the nested workflows after inlining](assets/docs/workflow_flatenned.svg)","ref":"workflows.html#nested-workflows"},{"type":"extras","title":"Operations","doc":"# Operations\n\nSkitter operations represent a reusable data processing step. An operation is\nresponsible for processing any data it receives, potentially emitting new data\nin the process. Operations receive inputs on their _input ports_ and emit\noutput on their _output ports_. An operation is always combined with a\ndistribution strategy, which determines how the operation is distributed.\n\nOperations are defined through the use of\n`Skitter.DSL.Operation.defoperation/3`. As a first example, consider the\nfollowing operation, which adds 1 to each data record it receives.\n\n```\ndefoperation Increment, in: value, out: incremented, strategy: Skitter.BIS.ImmutableLocal do\n  defcb react(value) do\n    value + 1 ~> incremented\n  end\nend\n```\n\nThis definition consists of meta-information about the operation and a\n_callback definition_. The meta-information specifies that the operation can\naccept values from a single input stream: `value`, that it produces a single\noutput stream: `incremented`, and that its default strategy is\n`Skitter.BIS.ImmutableLocal`.\n\nThe `Increment` operation defines a single callback, `react`, which is called\nwhen a new value arrives on the `value` stream. The callback is called with the\nreceived value, adds 1 to it and _emits_ the result to the `incremented` output\nstream through the use of the [`~>`](`Skitter.DSL.Operation.~>/2`) operator.\n\n## (Default) strategies\n\nAt runtime, an operation is always paired with a distribution strategy, which\ndetermines how it behaves. This pairing can happen in two places:\n\n* An operation can specify a default distribution strategy which is used when\n  the operation is not paired with another distribution strategy in a workflow.\n* When an operation is embedded in a workflow as a\n  [node](`Skitter.DSL.Workflow.node/2`), a strategy can be specified through\n  the use of `with:`. This strategy always takes precedence over a default\n  distribution strategy.\n\nA distribution strategy specifies which callbacks an operation should\nimplement. For instance, `Skitter.BIS.ImmutableLocal` specifies that an\noperation must define a `react` callback and that this callback will be called\nfor every incoming data element. It is therefore recommended to always specify\na default distribution strategy when defining an operation.","ref":"operations.html"},{"type":"extras","title":"Selecting a strategy - Operations","doc":"The appropriate strategy to use for an operation is determined by the\nproperties of the operation. For instance, if an operation is entirely\nstateless, `Skitter.BIS.ImmutableLocal` is an appropriate choice. If an\noperation needs to manage some state which is partitioned over several keys,\n`Skitter.BIS.KeyedState` may be the right choice. If none of the built-in\nstrategies of Skitter are appropriate for the operation you are writing, it is\npossible to define your own, as we discuss in the next section.","ref":"operations.html#selecting-a-strategy"},{"type":"extras","title":"Steps to write an operation - Operations","doc":"To define an operation, follow these steps:\n\n1. Figure out the properties of the operation.\n2. Based on these properties, select an appropriate distribution strategy.\n3. Read the documentation of the distribution strategy and determine which\n   callbacks your operation needs to implement.\n   - If you need to write your own distribution strategy, determine the\n     interface between the strategy and operation.\n4. Write the operation definition by defining the appropriate callbacks in\n   `Skitter.DSL.Operation.defoperation/3`.","ref":"operations.html#steps-to-write-an-operation"},{"type":"extras","title":"Dealing with state - Operations","doc":"Often, operations needs to manage some form of state. As an example, let's write\na `Count` operation which counts the amount of times it sees each value:\n\n```\ndefoperation Count, in: value, out: seen, strategy: Skitter.BIS.KeyedState do\n  initial_state 0\n\n  defcb key(value), do: value\n\n  defcb react(value) do\n    state() <~ state() + 1\n    {value, state()} ~> seen\n  end\nend\n```\n\nThe definition of this operation is similar to the definition of `Increment`.\nHowever, since this operation needs to manage state, we use the\n`Skitter.BIS.KeyedState` strategy. This strategy separates the incoming data\nelements based on some key, and maintains a state for each key. When a new data\nelement arrives, the `react` callback is called with the received value and\nwith the state of that key as its state. Inside react, this state can be\naccessed through the use of [`state()`](`Skitter.DSL.Operation.state/0`) and\nupdated through the use of [`state() <~`](`Skitter.DSL.Operation.<~/2`).\n\nThe operation also defines its initial state by defining a callback named\n`initial_state`, which accepts no arguments. As a syntactic convenience,\n[`initial_state`](`Skitter.DSL.Operation.initial_state/1`) can be used to\ndefine its initial state.\n\nFinally, the `Skitter.BIS.KeyedState` strategy requires that operations define\na `key` callback which obtains a key for an incoming data record. In the case\nof `Count`, the incoming data record is used as key.","ref":"operations.html#dealing-with-state"},{"type":"extras","title":"Complex state - Operations","doc":"Sometimes, operations maintain a complex state, defined as a struct with\nseveral fields. The operation DSL defines\n[`state_struct/1`](`Skitter.DSL.Operation.state_struct/1`),\n[`<~`](`Skitter.DSL.Operation.<~/2`) and\n[`~f()`](`Skitter.DSL.Operation.sigil_f/2`) to facilitate this:\n\n```\ndefoperation Average, in: value, out: average, strategy: Skitter.BIS.KeyedState do\n  state_struct total: 0, count: 0\n\n  defcb react(val) do\n    count <~ ~f{count} + 1\n    total <~ ~f{total} + val\n    ~f{total} / ~f{count} ~> average\n  end\nend\n```\n\nSee `Skitter.DSL.Operation.state_struct/1` for more information.","ref":"operations.html#complex-state"},{"type":"extras","title":"Other features - Operations","doc":"","ref":"operations.html#other-features"},{"type":"extras","title":"Emitting values - Operations","doc":"The [`~>`](`Skitter.DSL.Operation.~>/2`) operator is used to publish data to an\nout port inside a callback. The [`~>>`](`Skitter.DSL.Operation.~>>/2`) operator\ncan be used to emit several values at once. For instance, `[1, 2, 3] ~>> port`\nwill emit, `1`, `2` and `3` to port, one after the other.\nAs a special case, a (potentially infinite) Elixir stream can be passed to this\noperator. Every value in this stream will then be emitted to the output stream.\nIf the stream is infinite, the operation will not respond to new data elements,\nas it is busy emitting the values in the stream.","ref":"operations.html#emitting-values"},{"type":"extras","title":"Source operations - Operations","doc":"Sources are defined as operations in Skitter. To define a source, select a\ndistribution strategy which defines a source (such as\n`Skitter.BIS.StreamSource`) and implement the callbacks defined in its\ninterface.","ref":"operations.html#source-operations"},{"type":"extras","title":"Configuration data - Operations","doc":"Often, an operation is created with an immutable state, which remains\nunmodified throughout the lifetime of the operation. In Skitter, this immutable\nstate may be stored in the so-called _configuration_ data, which can be accessed\ninside a callback through the use of\n[`config()`](`Skitter.DSL.Operation.config/0`). Typically, strategies enable\noperations to implement a `conf` callback which defines the initial value of\nthis configuration data.\n\nAs an example, the `Skitter.BIO.Map` operation stores the function to execute\ninside its configuration data:\n\n```\ndefoperation Skitter.BIO.Map, in: _, out: _ strategy: Skitter.BIS.ImmutableLocal do\n  defcb conf(func), do: func\n  defcb react(arg), do: config().(arg) ~> _\nend\n```","ref":"operations.html#configuration-data"},{"type":"extras","title":"Accessing token information - Operations","doc":"A value received on an operation's input stream is wrapped inside a so-called\n_token_, which contains additional information about the received data, such\nas the port it was received on. If the strategy does not modify this token,\nthis information can be accessed inside a callback. For instance,\n[`port_of/1`](`Skitter.DSL.Operation.port_of/1`) can be used to obtain the port\na data record was received on.\n\nOperations can also add information to such a token through the use of\n[`extend_meta/3`](`Skitter.DSL.Operation.extend_meta/3`),\n[`inherit_meta/2`](`Skitter.DSL.Operation.inherit_meta/2`), or the functions\ndefined in `Skitter.Token`. Afterwards, this data can be obtained through the\nuse of [`meta_of/1`](`Skitter.DSL.Operation.meta_of/1`).","ref":"operations.html#accessing-token-information"},{"type":"extras","title":"Distribution Strategies","doc":"# Distribution Strategies","ref":"strategies.html"},{"type":"extras","title":"Deployment","doc":"# Deployment\n\nOnce you have a working stream processing application, you may wish to test it\nas a distributed system on your machine before distributing it over a cluster.\nThis guide documents how Skitter supports both of these scenarios.\n\n> #### Elixir, mix and releases {:.info}\n> Elixir offers several ways to run applications: during development,\n> developers compile and run their applications using `mix`, the elixir build\n> tool and `iex`, the interactive elixir shell. In order to run an application\n> in production, _releases_ (`mix release`) are commonly used.\n>\n> Skitter follows this convention. `mix` and `iex` are used to run an application\n> during development, while releases are built and used to deploy the application\n> over a cluster.","ref":"deployment.html"},{"type":"extras","title":"Development in local mode - Deployment","doc":"While writing an application, developers typically use `iex` to experiment with\ntheir application. Calling `iex -S mix` starts the interactive elixir shell and\nloads the current project into `iex`, enabling users to access the modules of\ntheir project and its dependencies. `mix run --no-halt` can be used to start\nthe project without starting an interactive shell.\n\nWhen a Skitter application (i.e. an application which uses Skitter as a\ndependency) is started using `iex -S mix` or `mix run`, Skitter will start in\nits so-called _local_ mode. In this mode, Skitter acts as both a coordinator\nand a runner for your distributed stream processing application, which is\nuseful for development.","ref":"deployment.html#development-in-local-mode"},{"type":"extras","title":"Simulating a distributed system - Deployment","doc":"Eventually, it becomes useful to simulate a distributed environment on your\nlocal machine. This can done through the use of `mix skitter.worker` and\n`mix skitter.master`. The former starts a worker node, which executes\ncomputations to perform, while the latter starts a master node, which divides\nwork among the workers. Setting up a working Skitter system requires one master\nnode and at least one worker node. Additionally, both nodes need to be\nconnected to one another.\n\nIn order to enable an Elixir node to connect to other nodes, it needs to be\nprovided with a name, this can be done through the use of the `--sname` option,\nwhich can be passed to the `elixir` or the `iex` command. For instance, to\nstart a worker named `worker`:\n\n<!-- tabs-open -->","ref":"deployment.html#simulating-a-distributed-system"},{"type":"extras","title":"iex - Deployment","doc":"```shell\niex --sname worker -S mix skitter.worker\n```","ref":"deployment.html#iex"},{"type":"extras","title":"elixir - Deployment","doc":"```shell\nelixir --sname worker -S mix skitter.worker\n```\n<!-- tabs-close -->\n\nAfterwards, you can start a master node in a different terminal window. You\nneed to provide the name of the worker node, suffixed with the hostname of your\ncomputer to enable both to work together:\n\n<!-- tabs-open -->","ref":"deployment.html#elixir"},{"type":"extras","title":"iex - Deployment","doc":"```shell\niex --sname master -S mix skitter.master worker@ \n```","ref":"deployment.html#iex"},{"type":"extras","title":"elixir - Deployment","doc":"```shell\nelixir --sname master -S mix skitter.master worker@ \n```\n<!-- tabs-close -->\n\nOnce the master is connected to the worker, it will deploy its workflow.\n\n<!-- tabs-open -->","ref":"deployment.html#elixir"},{"type":"extras","title":"master output - Deployment","doc":"```shell\n[20:05:21.629][info] ⬡⬢⬡⬢ Skitter v0.7.0 started in master mode\n[20:05:21.631][info] Reachable at `master@silverlode`\n[20:05:21.656][info] Connected to `worker@silverlode`, tags: []\n[20:05:21.667][info] Deploying &HelloSkitter.workflow/0\n```","ref":"deployment.html#master-output"},{"type":"extras","title":"worker output - Deployment","doc":"```shell\n[20:05:04.573][info] ⬡⬢⬡⬢ Skitter v0.7.0 started in worker mode\n[20:05:04.575][info] Reachable at `worker@silverlode`\n[20:05:21.656][info] Connected to master: `master@silverlode`\n{\"Skitter\", 1}\n{\"Hello\", 1}\n{\"Hello\", 2}\n{\"World!\", 1}\n```\n<!-- tabs-close -->\n\nYou can create as many worker nodes you like:\n\n<!-- tabs-open -->","ref":"deployment.html#worker-output"},{"type":"extras","title":"master output - Deployment","doc":"```shell\n$ elixir --sname master -S mix skitter.master worker1@silverlode worker2@silverlode\n[20:08:05.314][info] ⬡⬢⬡⬢ Skitter v0.7.0 started in master mode\n[20:08:05.316][info] Reachable at `master@silverlode`\n[20:08:05.341][info] Connected to `worker2@silverlode`, tags: []\n[20:08:05.342][info] Connected to `worker1@silverlode`, tags: []\n[20:08:05.354][info] Deploying &HelloSkitter.workflow/0\n```","ref":"deployment.html#master-output"},{"type":"extras","title":"worker1 output - Deployment","doc":"```shell\n$ elixir --sname worker1 -S mix skitter.worker\n[20:07:38.046][info] ⬡⬢⬡⬢ Skitter v0.7.0 started in worker mode\n[20:07:38.049][info] Reachable at `worker1@silverlode`\n[20:08:05.342][info] Connected to master: `master@silverlode`\n{\"Hello\", 2}\n{\"Hello\", 1}\n```","ref":"deployment.html#worker1-output"},{"type":"extras","title":"worker2 output - Deployment","doc":"```shell\n$ elixir --sname worker2 -S mix skitter.worker\n[20:07:50.918][info] ⬡⬢⬡⬢ Skitter v0.7.0 started in worker mode\n[20:07:50.920][info] Reachable at `worker2@silverlode`\n[20:08:05.342][info] Connected to master: `master@silverlode`\n{\"World!\", 1}\n{\"Skitter\", 1}\n```\n<!-- tabs-close -->","ref":"deployment.html#worker2-output"},{"type":"extras","title":"Deploying an application over a cluster - Deployment","doc":"While `mix skitter.worker` and `mix skitter.master` could be used to distribute\nan application over a cluster, Skitter instead uses _releases_ (`mix release`)\nfor this purpose. This is done for several reasons:\n\n- Releases are self-contained. This makes it easier to deploy an application\n  along with the Skitter runtime and its dependencies over a cluster.\n- Releases are somewhat optimized: unused modules are pruned and modules are\n  preloaded when the erlang vm is started.\n- Releases include management scripts, including the skitter deploy script,\n  which facilitates the distribution of a Skitter application over a cluster.\n\nA Skitter application is thus deployed over a cluster by:\n\n1. Compiling it into a release\n2. Using the skitter deploy script to start the distributed system","ref":"deployment.html#deploying-an-application-over-a-cluster"},{"type":"extras","title":"Creating a release - Deployment","doc":"A release is simply created by running `mix release` at the root of your\nproject (i.e. in the directory that has the `mix.exs` file).\n\nUsing the `hello_skitter` project discussed in the\n[up and running](up_and_running.html) guide, this looks as follows:\n\n```shell\n$ mix release\n \n* creating _release/releases/0.1.0/vm.args\n* creating _release/releases/0.1.0/remote.vm.args\n* creating _release/releases/0.1.0/env.sh\n* creating _release/bin/skitter\n```\n\nThe `_release` directory now contains a bunch of files which make up your\nrelease. In this guide, we will mainly interact with the `_release/bin/skitter`\nfile, which is a script that is used to deploy your application over a cluster.\n\n> #### Releases without `mix skitter.new` {:.info}\n>\n> `mix skitter.new` automatically configures the generated project for the easy\n> generation of releases. If you did not use `mix skitter.new`, or if you need\n> to adjust the project configuration for some other reasons, it is important\n> to ensure skitter can still customize the generated release. This is done by\n> adding `Skitter.Release.step/1` to the `:steps` used to build a release.\n> Please refer to the `Skitter.Release` documentation for more information.\n>\n> Additionally, you should ensure your releases are built in `:prod` mode. Once\n> again, this is the case when your project was created with `mix skitter.new`.\n\n> #### Copying releases to other machines {:.info}\n>\n> Releases can be copied to other machines as long as the target machine runs\n> the same operating system distribution as the machine which ran the\n> `mix release` command. The target machine does not need to have Erlang/Elixir\n> installed, as the Erlang and Elixir runtime system are bundled with the\n> release.\n>\n> Thus, you can compile your release on the master node of a cluster and copy\n> it over to the worker nodes.","ref":"deployment.html#creating-a-release"},{"type":"extras","title":"The skitter script - Deployment","doc":"The `skitter` script stored in `_release/bin/` can be used to start a\nstandalone Skitter master or worker node, or to deploy several workers and a\nsingle master of a cluster. The table below lists the supported commands; you\ncan get more information by running `_release/bin/skitter help`.\n\nCommand | Description\n------- | -----------\n`deploy` | Deploy a Skitter system over a set of nodes\n`local` | Manage a local Skitter runtime\n`worker` | Manage a Skitter worker runtime\n`master` | Manage a Skitter master runtime\n`help` | Get help for a given command\n`--application-version` | Get the version of the release application\n`--skitter-version` | Get the version of skitter included with this release\n\n`./bin/skitter local`, `./bin/skitter worker` and `./bin/skitter master` can be\nused to start local, worker or master runtimes respectively. Thus, you can\nmanually spawn a distributed system to test the release, similar to how it is\ndone with `mix`.\n\n<!-- tabs-open -->","ref":"deployment.html#the-skitter-script"},{"type":"extras","title":"master output - Deployment","doc":"```shell\n$ ./_release/bin/skitter master start worker1@silverlode worker2@silverlode\n[17:39:08.833][info] ⬡⬢⬡⬢ Skitter v0.7.0 started in master mode\n[17:39:08.833][info] Reachable at `skitter_master@silverlode`\n[17:39:08.997][info] Connected to `worker2@silverlode`, tags: []\n[17:39:08.998][info] Connected to `worker1@silverlode`, tags: []\n[17:39:08.999][info] Deploying &HelloSkitter.workflow/0\n```","ref":"deployment.html#master-output"},{"type":"extras","title":"worker1 output - Deployment","doc":"```shell\n$ ./_release/bin/skitter worker start --name worker1\n[17:38:40.608][info] ⬡⬢⬡⬢ Skitter v0.7.0 started in worker mode\n[17:38:40.608][info] Reachable at `worker1@silverlode`\n[17:39:08.999][info] Connected to master: `skitter_master@silverlode`\n{\"Hello\", 1}\n{\"Hello\", 2}\n```","ref":"deployment.html#worker1-output"},{"type":"extras","title":"worker2 output - Deployment","doc":"```shell\n$ ./_release/bin/skitter worker start --name worker2\n[17:38:51.599][info] ⬡⬢⬡⬢ Skitter v0.7.0 started in worker mode\n[17:38:51.599][info] Reachable at `worker2@silverlode`\n[17:39:08.997][info] Connected to master: `skitter_master@silverlode`\n{\"Skitter\", 1}\n{\"World!\", 1}\n```\n<!-- tabs-close -->\n\nPlease refer to the documentation returned by running\n`./_release/bin/skitter help worker` and `./_release/bin/skitter help master`\nfor more information.","ref":"deployment.html#worker2-output"},{"type":"extras","title":"skitter deploy - Deployment","doc":"Manually spawning master and worker nodes on a potentially large cluster is\ntedious. Therefore, the skitter script includes a `deploy` command, which deploys\nan application over a cluster, starting the required worker and master nodes as\nneeded.\n\nAt its simplest, you provide the script with the hostnames of the various\nworker machines of your cluster; the script will then ssh to each node and\nstart a worker. Once this is done, it will start a master node on the node\nwhere the deploy script was run; the master node will attempt to connect to\nevery worker node.\n\nOn a cluster with three nodes named `isabelle-a`, `isabelle-b` and\n`isabelle-c`, the output of the deploy script looks as follows when used\non the `hello_skitter` project.\n\n```shell\n$ ./_release/bin/skitter deploy isabelle-a isabelle-b isabelle-c\n⬡⬢⬡⬢ Skitter deploy 0.7.0\n> workers:  isabelle-a isabelle-b isabelle-c\n\n* starting worker on isabelle-a ✓\n* starting worker on isabelle-b ✓\n* starting worker on isabelle-c ✓\n\n⧖ sleeping 10 second(s) while workers start.\n✓ finished deployment, starting master.\n[20:58:29.950][info] ⬡⬢⬡⬢ Skitter v0.7.0 started in master mode\n[20:58:29.950][info] Reachable at `skitter_master@isabelle`\n[20:58:29.958][info] Connected to `skitter_worker@isabelle-a`, tags: []\n[20:58:29.959][info] Connected to `skitter_worker@isabelle-c`, tags: []\n[20:58:29.960][info] Connected to `skitter_worker@isabelle-b`, tags: []\n[20:58:29.960][info] Deploying &HelloSkitter.workflow/0\n```\n\nWhen using the deploy script, the cluster is set up to ensure the worker\nnodes quit if the master node quits. If any of the worker nodes cannot\nbe spawned, the deploy script stops the created workers and exits with\nan error.\n\nThe behaviour of the script can be tweaked as needed; please run\n`./_release/bin/skitter help deploy` for more information.\n\nThe deploy script makes the following assumptions:\n- The worker hosts are reachable over `ssh` without the need for a password\n- The release is available at the same location on all worker nodes.\n  - By default, it is assumed this location is the same as the directory from\n    which the release was started. When this is not the case, the\n    `--working-dir` option can be used. See `skitter help deploy` for more\n    information.","ref":"deployment.html#skitter-deploy"},{"type":"extras","title":"Configuration","doc":"# Configuration\n\nSkitter aims to be usable out of the box without any additional configuration.\nRunning the Skitter deploy script with the appropriate options should be enough\nto deploy a stream processing application over a cluster. Nevertheless, it may\nbe needed to tweak the behavior of Skitter. This page details the various\nconfiguration options Skitter offers to modify its behavior.\n\n> #### Configuring Elixir applications {:.tip}\n>\n> Elixir offers several options to configure the behavior of an application.\n> This can sometimes make it difficult to figure out where configuration should\n> go. The following list provides a quick overview of the various ways in which\n> an application can be configured.\n>\n> * `config/config.exs`: compile-time configuration, created by default by\n>    `mix skitter.new`. See `Config` for more information.\n> * `config/runtime.exs`: runtime configuration. Evaluated every time before\n>    the application is started. Not created by `mix skitter.new`.\n>\n> Additionally, Skitter can be configured by passing command-line parameters to\n> `mix skitter.worker`, `mix skitter.master`, `skitter worker`,\n> `skitter master`, `skitter local` or `skitter deploy`.\n\nA Skitter runtime is always started in a _mode_ (i.e., worker, master or local).\nSome configuration options are only useful when Skitter is running in a certain\nmode. Skitter ignores any configuration options which are not relevant for the\nmode it is started in.\n\nThe available configuration options are documented below.","ref":"configuration.html"},{"type":"extras","title":"deploy - Configuration","doc":"> #### Summary {:.neutral}\n>\n> Deploy the specified workflow after starting Skitter.\n>\n> - _master_ or _local_ mode\n> - Default: `nil`\n> - `config/config.exs`: `config :skitter, deploy:  `\n> - `config/runtime.exs`: `config :skitter, deploy:  `\n> - `mix skitter.master  --deploy  ` or `mix skitter.master -d  `\n> - `skitter master  --deploy  ` or `skitter master -d  `\n> - `skitter deploy  --deploy  ` or `skitter deploy -d  `\n\nThis configuration option is used to specify a workflow which will be deployed\nover the cluster after the Skitter runtime has started. There are two ways to\nconfigure this option:\n\n- Setting the `deploy` option in `config/config.exs` or `config/runtime.exs` to\n  a 0-arity function. This function should return a workflow which will be\n  deployed by Skitter.\n  - This is the case if the project was generated using `mix skitter.new`.\n- An expression can be passed as an argument to the `--deploy` flag. This\n  expression will be evaluated using `Code.eval_string/3`. The resulting value\n  should be a workflow, which will be deployed after the Skitter runtime has\n  started.","ref":"configuration.html#deploy"},{"type":"extras","title":"logs - Configuration","doc":"> #### Summary {:.neutral}\n>\n> Write logs to files\n>\n> - _worker_ and _master_ mode\n> - Default: `true`\n> - Only for release builds\n> - `skitter worker --no-log`\n> - `skitter master --no-log`\n> - `skitter deploy --no-log`\n\nSkitter writes messages to Elixir's `Logger`. Skitter applications which were\ncreated using `mix skitter.new` write their log messages to standard error.\nThis behavior (and other logger settings) can be modified by configuring the\nlogger in `config/config.exs`.\n\nAdditionally, skitter release builds are configured to write their log output\nto a file (`logs/ .log`) if they are started in master or worker mode.\nThis behavior can be disabled by passing the `--no-log` option to the skitter\nscript.","ref":"configuration.html#logs"},{"type":"extras","title":"workers - Configuration","doc":"> #### Summary {:.neutral}\n>\n> A list of workers to which the master will attempt to connect.\n>\n> - _master_ mode\n> - Default: `[]`\n> - `mix skitter.master    `\n> - `skitter master    `\n> - `skitter master --worker-file  ` or `skitter master -f  `\n> - `skitter deploy    `\n> - `skitter deploy --worker-file  ` or `skitter deploy -f  `\n\nA list of workers to which the master will attempt to connect. If connecting to\nany of these workers fail, the master will shut down with an error.\n\nA path to a file may also be provided to `skitter master` or `skitter deploy`.\nThis file must contain a worker address on each line. These workers will be\nadded to the list of workers.","ref":"configuration.html#workers"},{"type":"extras","title":"shutdown_with_workers - Configuration","doc":"> #### Summary {:.neutral}\n>\n> Shut down the master when any connected worker shuts down.\n>\n> - _master_ mode\n> - Default: false\n> - `skitter master --shutdown-with-workers`\n> - `skitter deploy --shutdown-with-workers`\n\nThis option is useful to ensure a single crashed worker shuts down all\nconnected Skitter runtimes.","ref":"configuration.html#shutdown_with_workers"},{"type":"extras","title":"master - Configuration","doc":"> #### Summary {:.neutral}\n>\n> A master to connect to.\n>\n> - _worker_ mode\n> - Default: `nil`\n> - `mix skitter.worker  `\n> - `skitter worker  `\n\nAfter starting, the worker will attempt to connect to the master node. If the\nconnection fails, the worker will log a warning but stays alive.","ref":"configuration.html#master"},{"type":"extras","title":"shutdown_with_workers - Configuration","doc":"> #### Summary {:.neutral}\n>\n> Shut down the worker if the master it is connected to shuts down.\n>\n> - _worker_ mode\n> - Default: `true`\n> - ` mix skitter.worker --no-shutdown-with-master`\n> - `skitter worker --no-shutdown-with-master`\n> - `skitter deploy --no-shutdown-with-master`","ref":"configuration.html#shutdown_with_workers"},{"type":"extras","title":"tags - Configuration","doc":"> #### Summary {:.neutral}\n>\n> Add the specified tags to the worker\n>\n> - _worker_ mode\n> - Default: `true`\n> - `mix skitter.worker -t   -t  `\n> - `skitter worker -t   -t  `\n> - `skitter deploy --worker-file  ` or `skitter deploy -f  `\n\nA list of `t:Skitter.Remote.tag/0` which will be added to the worker.\n\nThe `--worker-file` provides a special notation which can be used to add tags\nto workers.","ref":"configuration.html#tags"},{"type":"extras","title":"telemetry - Configuration","doc":"> #### Summary {:.neutral}\n>\n> Enable telemetry events.\n>\n> - All modes\n> - Default: `false`\n> - Must be set at compile-time\n> - `config/config.exs`: `config :skitter, telemetry:  `\n\nSkitter can optionally emit telemetry events through the use of the `telemetry`\npackage. This option determines whether these events are emitted or not. If\nthis option is set to `false` (the default), all telemetry code is purged at\ncompile time. Therefore, this option can *only* be adjusted in\n`config/config.exs`. More information about telemetry can be found on the\n[telemetry page](telemetry.html).","ref":"configuration.html#telemetry"},{"type":"extras","title":"Operators","doc":"# Operators\n\nMany stream processing frameworks (e.g. Spark, Flink, …) use an operator-based\nprogramming model where applications are expressed by chaining calls to\nwell-known operations such as `map`, `join`, `reduce`, and others. In Skitter,\nhowever, applications are expressed by combining operations defined elsewhere.\n\nNevertheless, the well-known operations described above can easily be\nimplemented in Skitter. Skitter provides the following operations out of the box:\n\n* `Skitter.BIO.Filter`\n* `Skitter.BIO.FlatMap`\n* `Skitter.BIO.KeyedReduce`\n* `Skitter.BIO.Map`\n* `Skitter.BIO.MessageSource`\n* `Skitter.BIO.Print`\n* `Skitter.BIO.Send`\n* `Skitter.BIO.StreamSource`\n* `Skitter.BIO.TCPSource`\n\nThese operations can be used in a workflow definition:\n```\nworkflow do\n  node(Skitter.BIO.StreamSource, args: ~w(Hello Skitter Hello World!))\n  ~> node(Skitter.BIO.FlatMap, args: &String.split/1)\n  ~> node(Skitter.BIO.KeyedReduce, args: {\n    fn word -> word end,\n    fn word, ctr -> {ctr + 1, {word, ctr + 1}} end,\n    0\n  })\n  ~> node(Skitter.BIO.Print)\nend\n```\nHowever, it quickly becomes tedious to write applications like this. Therefore,\nthe `Skitter.BIO` module defines several macros which provide shorthands for\nusing these built-in operations. The\n[Skitter workflow DSL](`Skitter.DSL.Workflow.workflow/2`) imports these\nshorthands implicitly. Thus the application above can be written as follows:\n```\nworkflow do\n  stream_source(~w(Hello Skitter Hello World!))\n  ~> flat_map(&String.split/1)\n  ~> keyed_reduce(fn word -> word end, fn word, ctr -> {ctr + 1, {word, ctr + 1}} end, 0)\n  ~> print()\nend\n```\nIn Skitter, these macros are called _operators_.","ref":"operators.html"},{"type":"extras","title":"Defining Custom Operators - Operators","doc":"It is possible to define your own operators which can then be imported into the\nworkflow DSL.\n\nFor instance, `Skitter.BIO.map/2` is defined as follows:\n\n```\ndefmacro map(func, opts \\\\ []) do\n  opts = [args: func] ++ opts\n  quote(do: node(Skitter.BIO.Map, unquote(opts)))\nend\n```\n\nThis macro accepts two arguments: the function which will be mapped over the\nincoming data and any other options (i.e. the `as:` and `with:` options\naccepted by `Skitter.DSL.Workflow.node/2`). Based on these, it creates a call\nto `Skitter.DSL.Workflow.node/2`, with the function argument merged into the\nnode's option list. This call to `node` must be quoted, as it will be inserted\ninto the workflow DSL.\n\nThus, an operator definition must do the following:\n- It must be defined as a macro.\n- It must return a quoted call to `Skitter.DSL.Workflow.node/2`.\n- It must accept optional options, which are passed as the second argument to\n  `Skitter.DSL.Workflow.node/2`.\n\nOperators defined this way can then be imported into the workflow DSL using\n`import/2`.\n\nTo wrap up, let's show a complete example of defining and using a custom\noperator. We will define an operator which uses `Skitter.BIO.FlatMap` and embed\nit in the workflow shown above.\n\nFirst, we define the operator in a module:\n\n```\ndefmodule MyCustomOperators do\n  defmacro my_flat_map(func, opts \\\\ []) do\n    quote(do: node(Skitter.BIO.FlatMap, unquote([args: func] ++ opts)))\n  end\nend\n```\n\nAfter, we define a workflow importing the module after which we can use the\noperator:\n\n```\nworkflow do\n  import MyCustomOperators\n\n  stream_source(~w(Hello Skitter Hello World!))\n  ~> my_flat_map(&String.split/1)\n  ~> keyed_reduce(fn word -> word end, fn word, ctr -> {ctr + 1, {word, ctr + 1}} end, 0)\n  ~> print()\nend\n```","ref":"operators.html#defining-custom-operators"},{"type":"extras","title":"Telemetry","doc":"# Telemetry\n\n> #### Info {:.neutral}\n>\n> This page is only relevant for those who wish to intercept the telemetry\n> events produced by Skitter. It is not relevant if you wish to write and\n> deploy a Skitter application.\n\nSkitter emits various telemetry events which can be used to introspect the\ncurrent state of the runtime system. This page details the various events\nproduced by the Skitter runtime system, along with the meta-information passed\nwith these events. Note that these events are only produced when `telemetry` is\nset to `true` in the [Skitter application environment](configuration.html).","ref":"telemetry.html"},{"type":"extras","title":"Writing telemetry handlers - Telemetry","doc":"> #### Telemetry handlers {:.warning}\n>\n> Skitter does not define or register any handlers for the emitted telemetry\n> events. It is up to the user of Skitter to define telemetry handlers or to\n> use an external package which does.\n\nPlease refer to the [telemetry docs](`e:telemetry:readme.html`) for information\non how to write telemetry handlers. In a nutshell, Skitter emits several\ntelemetry _events_, which can be _attached_ to _handlers_, written by the user,\nwhich will receive the event and process it in some way.\n\nWhen a telemetry event occurs, registered handlers for the event type are\ncalled in the (Elixir) process where the event occurred. Therefore, registered\nhandlers may be called on any process on any remote cluster node. It is up to\nthe user to gather the data in a central location (e.g. on the master node),\nbefore it is processed.\n\nTelemetry events emitted by Skitter include a lot of information which may be\nquite large. For instance, many events include a `t:Skitter.Strategy.context/0`\nstruct, which contains a reference to the deployment data, which may be quite\nlarge. Developers who write event handlers which forward events to other\nprocesses should keep this in mind and only forward the meta information they\nrequire. This is especially relevant when data may be sent across the network.","ref":"telemetry.html#writing-telemetry-handlers"},{"type":"extras","title":"Associating events with application information - Telemetry","doc":"It is often needed to associate a telemetry event with a deployed workflow, or\nwith a node in a workflow. However, the telemetry events emitted by Skitter do\nnot contain this information. Instead, they contain a\n`t:Skitter.Strategy.context/0`. The functions in `Skitter.Runtime` may be used\nto obtain the workflow or workflow node associated with such a context.","ref":"telemetry.html#associating-events-with-application-information"},{"type":"extras","title":"Telemetry events - Telemetry","doc":"","ref":"telemetry.html#telemetry-events"},{"type":"extras","title":"Wrapped events - Telemetry","doc":"The events described in this section are wrapped in a call to\n`:telemetry.span/3`. This means that a `:start`, `:stop`, and `:exception`\nhandler may be defined for each of these events. For instance, to handle the\n`[:skitter, :foo]` event, handlers for `[:skitter, :foo, :start]`,\n`[:skitter, :foo, :stop]` and `[:skitter, :foo, :exception]` may need to be\ndefined. Please refer to the `:telemetry.span/3` documentation for more\ninformation.\n\nThe listed fields are added to both `:start` and `:stop` events. Furthermore,\na `:result` field is added to the meta-information of every `:stop` event,\nwhich contains the result of the wrapped code.\n\n#### Hooks\n\n* `[:skitter, :hook, :deploy]`: Emitted when the\n  `c:Skitter.Strategy.Operation.deploy/2` hook of a strategy is called.\n  * `:context`: The context passed to `c:Skitter.Strategy.Operation.deploy/2`\n  * `:result` (only for the `:stop` event): the return value of the hook. This\n    data will be stored inside the strategy's deployment.\n* `[:skitter, :hook, deliver]`: Emitted when the `deliver` hook of a strategy is\n  * `:context`, `token`: the arguments passed to\n    `c:Skitter.Strategy.Operation.deliver/2`.\n  * `:pid`: The `t:pid/0` of the process calling the hook. Note that this hook\n    is called from within a worker of the strategy emitting the data, so this\n    pid will not refer to a worker of the operation which should receive the\n    data. Instead, it will refer to a worker of its predecessor.\n  * `:result` (only for the `:stop` event). The result of the\n    `c:Skitter.Strategy.Operation.deliver/2` hook is not used by the runtime\n    system, so you should not use this.\n* `[:skitter, :hook, process]`: Emitted when the\n  `c:Skitter.Strategy.Operation.process/4` hook of a strategy is called.\n  * `:context`, `:message`, `:state`, `:role`: the arguments passed to\n    `c:Skitter.Strategy.Operation.process/4`.\n  * `:pid`: The `t:pid/0` of the worker calling the hook.\n  * `:result` (only for the `:stop` event): the return value of the hook. This\n    data will be used as the new state of the worker which called the hook.\n\n#### Operation callbacks\n\n* `[:skitter, :operation, :call]`: Emitted when an operation callback is called.\n  * `:pid`: The `t:pid/0` of the worker calling the callback.\n  * `:operation`, `:name`, `:state`, `:config`, `:args`: The arguments passed\n    `Skitter.Operation.call/5`.","ref":"telemetry.html#wrapped-events"},{"type":"extras","title":"Unwrapped events - Telemetry","doc":"The events described in this section are not wrapped and are emitted by\n`:telemetry.execute/3`.\n\n#### Runtime\n\n* `[:skitter, :worker, :init]`: Emitted when a new worker is initialized. Note\n  that, when a workflow is deployed, the workers are only initialized after the\n  `deploy` hook of each node in the workflow is called.\n  * `context`: The context the worker was deployed with.\n  * `state`: The initial state of the worker.\n  * `role`: The worker `t:Skitter.Worker.role/0`\n  * `pid`: The `t:pid/0` of the worker. Uniquely identifies the worker.\n* `[:skitter, :worker, :send]`: Emitted when a message is sent using\n  `Skitter.Worker.send/2`.\n  * `:from`: The `t:pid/0` of the worker sending the message.\n  * `:to`: The `t:pid/0` of the worker receiving the message.\n  * `:message`: The message being sent.\n* `[:skitter, :runtime, :emit]`: Emitted when a strategy emits data using\n  `Skitter.Strategy.Operation.emit/2`.\n  * `context`: The context of the hook emitting the data.\n  * `emit`: The emitted data: `t:Skitter.Operation.emit/0`.\n* `[:skitter, :runtime, :deploy]`: Emitted when a workflow is deployed. The\n  event is emitted after everything is ready, but before\n  `Skitter.Runtime.deploy/1` returns.\n  * `ref`: A unique reference to the deployed workflow.\n* `[:skitter, :runtime, :stop]`: Emitted when a deployed workflow is stopped\n  using `Skitter.Runtime.stop/1`. The event is emitted before any workers are\n  stopped. Telemetry events for `ref` received after this event should be\n  ignored, as no guarantees are made about the internal consistency of the\n  workflow.\n  * `ref`: The reference of the deployed workflow.\n\n#### Remote\n\n* `[:skitter, :remote, :up, :worker]`: Emitted when a worker runtime is\n  connected to a master runtime. This event can only occur on a skitter runtime\n  in `:master` mode.\n  * `remote`: The name of the remote runtime we connected to.\n  * `:tags`: The list of worker `t:Skitter.Remote.tag/0`\n* `[:skitter, :remote, :down, :worker]`: Emitted when we disconnect from a\n  worker. This event can only occur on a skitter runtime in `:master` mode.\n  * `remote`: The name of the remote runtime we disconnected from.\n  * `reason`: `:down` or `:remove`. In the first case, the remote runtime went\n    down, in the second case, the remote runtime was briefly connected to the\n    local runtime, after which it rejected the connection.","ref":"telemetry.html#unwrapped-events"}],"content_type":"text/markdown","producer":{"name":"ex_doc","version":[48,46,51,52,46,49]}}